import duckdb
import os
import pytest

pyarrow = pytest.importorskip("pyarrow")
np = pytest.importorskip("numpy")
pyarrow.parquet = pytest.importorskip("pyarrow.parquet")
pyarrow.dataset = pytest.importorskip("pyarrow.dataset")


class TestArrowDataset(object):
    def test_parallel_dataset(self, duckdb_cursor):
        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("PRAGMA threads=4")
        duckdb_conn.execute("PRAGMA verify_parallelism")

        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')

        userdata_parquet_dataset = pyarrow.dataset.dataset(
            [
                parquet_filename,
                parquet_filename,
                parquet_filename,
            ],
            format="parquet",
        )

        rel = duckdb_conn.from_arrow(userdata_parquet_dataset)

        assert (
            rel.filter("first_name=\'Jose\' and salary > 134708.82").aggregate('count(*)').execute().fetchone()[0] == 12
        )

    def test_parallel_dataset_register(self, duckdb_cursor):
        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("PRAGMA threads=4")
        duckdb_conn.execute("PRAGMA verify_parallelism")

        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')

        userdata_parquet_dataset = pyarrow.dataset.dataset(
            [
                parquet_filename,
                parquet_filename,
                parquet_filename,
            ],
            format="parquet",
        )

        rel = duckdb_conn.register("dataset", userdata_parquet_dataset)

        assert (
            duckdb_conn.execute(
                "Select count(*) from dataset where first_name = 'Jose' and salary > 134708.82"
            ).fetchone()[0]
            == 12
        )

    def test_parallel_dataset_roundtrip(self, duckdb_cursor):
        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("PRAGMA threads=4")
        duckdb_conn.execute("PRAGMA verify_parallelism")

        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')

        userdata_parquet_dataset = pyarrow.dataset.dataset(
            [
                parquet_filename,
                parquet_filename,
                parquet_filename,
            ],
            format="parquet",
        )

        rel = duckdb_conn.register("dataset", userdata_parquet_dataset)

        query = duckdb_conn.execute("SELECT * FROM dataset order by id")
        record_batch_reader = query.fetch_record_batch(2048)

        arrow_table = record_batch_reader.read_all()
        # reorder since order of rows isn't deterministic
        df = userdata_parquet_dataset.to_table().to_pandas().sort_values('id').reset_index(drop=True)
        # turn it into an arrow table
        arrow_table_2 = pyarrow.Table.from_pandas(df)
        result_1 = duckdb_conn.execute("select * from arrow_table order by all").fetchall()

        result_2 = duckdb_conn.execute("select * from arrow_table_2 order by all").fetchall()

        assert result_1 == result_2

    def test_ducktyping(self, duckdb_cursor):
        duckdb_conn = duckdb.connect()
        dataset = CustomDataset()
        query = duckdb_conn.execute("SELECT b FROM dataset WHERE a < 5")
        record_batch_reader = query.fetch_record_batch(2048)
        arrow_table = record_batch_reader.read_all()
        assert arrow_table.equals(CustomDataset.DATA[:5].select(['b']))


class CustomDataset(pyarrow.dataset.Dataset):
    # For testing duck-typing of dataset/scanner https://github.com/duckdb/duckdb/pull/5998
    SCHEMA = pyarrow.schema([pyarrow.field("a", pyarrow.int64(), True), pyarrow.field("b", pyarrow.float64(), True)])
    DATA = pyarrow.Table.from_arrays([pyarrow.array(range(100)), pyarrow.array(np.arange(100) * 1.0)], schema=SCHEMA)

    def __init__(self):
        pass

    def scanner(self, **kwargs):
        return CustomScanner(**kwargs)

    @property
    def schema(self):
        return CustomDataset.SCHEMA


class CustomScanner(pyarrow.dataset.Scanner):
    def __init__(self, filter=None, columns=None, **kwargs):
        self.filter = filter
        self.columns = columns
        self.kwargs = kwargs

    @property
    def projected_schema(self):
        if self.columns is None:
            return CustomDataset.SCHEMA
        else:
            return pyarrow.schema([f for f in CustomDataset.SCHEMA.fields if f.name in self.columns])

    def to_reader(self):
        return pyarrow.dataset.dataset(CustomDataset.DATA).scanner(filter=self.filter, columns=self.columns).to_reader()



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")

from duckdb.experimental.spark.sql.column import Column
from duckdb.experimental.spark.sql.functions import struct
from duckdb.experimental.spark.sql.types import Row
from duckdb.experimental.spark.errors import PySparkTypeError

import duckdb
import re


class TestSparkColumn(object):
    def test_struct_column(self, spark):
        df = spark.createDataFrame([Row(a=1, b=2, c=3, d=4)])

        df = df.withColumn('struct', struct(df.col0, df.col1))
        assert 'struct' in df
        new_col = df.schema['struct']
        assert 'col0' in new_col.dataType
        assert 'col1' in new_col.dataType

        with pytest.raises(
            PySparkTypeError, match=re.escape("[NOT_COLUMN] Argument `col` should be a Column, got str.")
        ):
            df = df.withColumn('struct', 'yes')



import pandas as pd
import duckdb
import datetime
import pytest


def run_checks(df):
    assert type(df['d'][0]) is datetime.date
    assert df['d'][0] == datetime.date(1992, 7, 30)
    assert pd.isnull(df['d'][1])


def test_date_as_datetime():
    con = duckdb.connect()
    con.execute("create table t (d date)")
    con.execute("insert into t values ('1992-07-30'), (NULL)")

    # Connection Methods
    run_checks(con.execute("Select * from t").df(date_as_object=True))
    run_checks(con.execute("Select * from t").fetchdf(date_as_object=True))
    run_checks(con.execute("Select * from t").fetch_df_chunk(date_as_object=True))
    run_checks(con.execute("Select * from t").fetch_df(date_as_object=True))

    # Relation Methods
    rel = con.table('t')
    run_checks(rel.df(date_as_object=True))
    run_checks(rel.to_df(date_as_object=True))

    # Result Methods
    run_checks(rel.query("t_1", "select * from t_1").df(date_as_object=True))



def vectorized(func):
    """
    Decorate a function with annotated function parameters, so DuckDB can infer that the function should be provided with pyarrow arrays and should expect pyarrow array(s) as output
    """
    from inspect import signature
    import types

    new_func = types.FunctionType(func.__code__, func.__globals__, func.__name__, func.__defaults__, func.__closure__)
    # Construct the annotations:
    import pyarrow as pa

    new_annotations = {}
    sig = signature(func)
    sig.parameters
    for param in sig.parameters:
        new_annotations[param] = pa.lib.ChunkedArray

    new_func.__annotations__ = new_annotations
    return new_func



from re import S
import duckdb
import os
import pytest
import tempfile
from conftest import pandas_supports_arrow_backend
import sys
from packaging.version import Version

pa = pytest.importorskip("pyarrow")
pq = pytest.importorskip("pyarrow.parquet")
ds = pytest.importorskip("pyarrow.dataset")
np = pytest.importorskip("numpy")
re = pytest.importorskip("re")


def create_pyarrow_pandas(rel):
    if not pandas_supports_arrow_backend():
        pytest.skip(reason="Pandas version doesn't support 'pyarrow' backend")
    return rel.df().convert_dtypes(dtype_backend='pyarrow')


def create_pyarrow_table(rel):
    return rel.arrow()


def create_pyarrow_dataset(rel):
    table = create_pyarrow_table(rel)
    return ds.dataset(table)


def numeric_operators(connection, data_type, tbl_name, create_table):
    connection.execute(
        f"""
        CREATE TABLE {tbl_name} (
            a {data_type},
            b {data_type},
            c {data_type}
        )
    """
    )
    connection.execute(
        f"""
        INSERT INTO {tbl_name} VALUES
            (1,1,1),
            (10,10,10),
            (100,10,100),
            (NULL,NULL,NULL)
    """
    )
    duck_tbl = connection.table(tbl_name)
    arrow_table = create_table(duck_tbl)

    # Try ==
    assert connection.execute("SELECT count(*) from arrow_table where a = 1").fetchone()[0] == 1
    # Try >
    assert connection.execute("SELECT count(*) from arrow_table where a > 1").fetchone()[0] == 2
    # Try >=
    assert connection.execute("SELECT count(*) from arrow_table where a >= 10").fetchone()[0] == 2
    # Try <
    assert connection.execute("SELECT count(*) from arrow_table where a < 10").fetchone()[0] == 1
    # Try <=
    assert connection.execute("SELECT count(*) from arrow_table where a <= 10").fetchone()[0] == 2

    # Try Is Null
    assert connection.execute("SELECT count(*) from arrow_table where a IS NULL").fetchone()[0] == 1
    # Try Is Not Null
    assert connection.execute("SELECT count(*) from arrow_table where a IS NOT NULL").fetchone()[0] == 3

    # Try And
    assert connection.execute("SELECT count(*) from arrow_table where a = 10 and b = 1").fetchone()[0] == 0
    assert (
        connection.execute("SELECT count(*) from arrow_table where a = 100 and b = 10 and c = 100").fetchone()[0] == 1
    )

    # Try Or
    assert connection.execute("SELECT count(*) from arrow_table where a = 100 or b = 1").fetchone()[0] == 2

    connection.execute("EXPLAIN SELECT count(*) from arrow_table where a = 100 or b = 1")
    print(connection.fetchall())


def numeric_check_or_pushdown(connection, tbl_name, create_table):
    duck_tbl = connection.table(tbl_name)
    arrow_table = create_table(duck_tbl)

    # Multiple column in the root OR node, don't push down
    query_res = connection.execute(
        """
        EXPLAIN SELECT * FROM arrow_table WHERE
            a = 1 OR b = 2 AND (a > 3 OR b < 5)
    """
    ).fetchall()
    match = re.search(".*ARROW_SCAN.*Filters:.*", query_res[0][1])
    assert not match

    # Single column in the root OR node
    query_res = connection.execute(
        """
        EXPLAIN SELECT * FROM arrow_table WHERE
            a = 1 OR a = 10
    """
    ).fetchall()
    match = re.search(".*ARROW_SCAN.*Filters: a=1 OR a=10.*|$", query_res[0][1])
    assert match

    # Single column + root OR node with AND
    query_res = connection.execute(
        """
        EXPLAIN SELECT * FROM arrow_table
            WHERE a = 1 OR (a > 3 AND a < 5)
    """
    ).fetchall()
    match = re.search(".*ARROW_SCAN.*Filters: a=1 OR a>3 AND a<5.*|$", query_res[0][1])
    assert match

    # Single column multiple ORs
    query_res = connection.execute("EXPLAIN SELECT * FROM arrow_table WHERE a=1 OR a>3 OR a<5").fetchall()
    match = re.search(".*ARROW_SCAN.*Filters: a=1 OR a>3 OR a<5.*|$", query_res[0][1])
    assert match

    # Testing not equal
    query_res = connection.execute("EXPLAIN SELECT * FROM arrow_table WHERE a!=1 OR a>3 OR a<2").fetchall()
    match = re.search(".*ARROW_SCAN.*Filters: a!=1 OR a>3 OR a<2.*|$", query_res[0][1])
    assert match

    # Multiple OR filters connected with ANDs
    query_res = connection.execute(
        "EXPLAIN SELECT * FROM arrow_table WHERE (a<2 OR a>3) AND (a=1 OR a=4) AND (b=1 OR b<5)"
    ).fetchall()
    match = re.search(".*ARROW_SCAN.*Filters: a<2 OR a>3 AND a=1|\n.*OR a=4.*\n.*b=2 OR b<5.*|$", query_res[0][1])
    assert match


def string_check_or_pushdown(connection, tbl_name, create_table):
    duck_tbl = connection.table(tbl_name)
    arrow_table = create_table(duck_tbl)

    # Check string zonemap
    query_res = connection.execute("EXPLAIN SELECT * FROM arrow_table WHERE a >= '1' OR a <= '10'").fetchall()
    match = re.search(".*ARROW_SCAN.*Filters: a>=1 OR a<=10.*|$", query_res[0][1])
    assert match

    # No support for OR with is null
    query_res = connection.execute("EXPLAIN SELECT * FROM arrow_table WHERE a IS NULL or a = '1'").fetchall()
    match = re.search(".*ARROW_SCAN.*Filters:.*", query_res[0][1])
    assert not match

    # No support for OR with is not null
    query_res = connection.execute("EXPLAIN SELECT * FROM arrow_table WHERE a IS NOT NULL OR a = '1'").fetchall()
    match = re.search(".*ARROW_SCAN.*Filters:.*", query_res[0][1])
    assert not match

    # OR with the like operator
    query_res = connection.execute("EXPLAIN SELECT * FROM arrow_table WHERE a = 1 OR a LIKE '10%'").fetchall()
    match = re.search(".*ARROW_SCAN.*Filters:.*", query_res[0][1])
    assert not match


class TestArrowFilterPushdown(object):
    @pytest.mark.parametrize(
        'data_type',
        [
            'TINYINT',
            'SMALLINT',
            'INTEGER',
            'BIGINT',
            'UTINYINT',
            'USMALLINT',
            'UINTEGER',
            'UBIGINT',
            'FLOAT',
            'DOUBLE',
            'HUGEINT',
            'DECIMAL(4,1)',
            'DECIMAL(9,1)',
            'DECIMAL(18,4)',
            'DECIMAL(30,12)',
        ],
    )
    @pytest.mark.parametrize('create_table', [create_pyarrow_pandas, create_pyarrow_table])
    def test_filter_pushdown_numeric(self, data_type, duckdb_cursor, create_table):
        tbl_name = "tbl"
        numeric_operators(duckdb_cursor, data_type, tbl_name, create_table)
        numeric_check_or_pushdown(duckdb_cursor, tbl_name, create_table)

    @pytest.mark.parametrize('create_table', [create_pyarrow_pandas, create_pyarrow_table])
    def test_filter_pushdown_varchar(self, duckdb_cursor, create_table):
        duckdb_cursor.execute(
            """
            CREATE TABLE test_varchar (
                a VARCHAR,
                b VARCHAR,
                c VARCHAR
            )
        """
        )
        duckdb_cursor.execute(
            """
            INSERT INTO test_varchar VALUES
                ('1','1','1'),
                ('10','10','10'),
                ('100','10','100'),
                (NULL, NULL, NULL)
        """
        )
        duck_tbl = duckdb_cursor.table("test_varchar")
        arrow_table = create_table(duck_tbl)

        # Try ==
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a = '1'").fetchone()[0] == 1
        # Try >
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a > '1'").fetchone()[0] == 2
        # Try >=
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a >= '10'").fetchone()[0] == 2
        # Try <
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a < '10'").fetchone()[0] == 1
        # Try <=
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a <= '10'").fetchone()[0] == 2

        # Try Is Null
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a IS NULL").fetchone()[0] == 1
        # Try Is Not Null
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a IS NOT NULL").fetchone()[0] == 3

        # Try And
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a = '10' and b = '1'").fetchone()[0] == 0
        assert (
            duckdb_cursor.execute(
                "SELECT count(*) from arrow_table where a = '100' and b = '10' and c = '100'"
            ).fetchone()[0]
            == 1
        )
        # Try Or
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a = '100' or b ='1'").fetchone()[0] == 2

        # More complex tests for OR pushed down on string
        string_check_or_pushdown(duckdb_cursor, "test_varchar", create_table)

    @pytest.mark.parametrize('create_table', [create_pyarrow_pandas, create_pyarrow_table])
    def test_filter_pushdown_bool(self, duckdb_cursor, create_table):
        duckdb_cursor.execute(
            """
            CREATE TABLE test_bool (
                a BOOL,
                b BOOL
            )
        """
        )
        duckdb_cursor.execute(
            """
            INSERT INTO test_bool VALUES
                (TRUE,TRUE),
                (TRUE,FALSE),
                (FALSE,TRUE),
                (NULL,NULL)
        """
        )
        duck_tbl = duckdb_cursor.table("test_bool")
        arrow_table = create_table(duck_tbl)

        # Try ==
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a = True").fetchone()[0] == 2

        # Try Is Null
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a IS NULL").fetchone()[0] == 1
        # Try Is Not Null
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a IS NOT NULL").fetchone()[0] == 3

        # Try And
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a= True and b = True").fetchone()[0] == 1
        # Try Or
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a = True or b = True").fetchone()[0] == 3

    @pytest.mark.parametrize('create_table', [create_pyarrow_pandas, create_pyarrow_table])
    def test_filter_pushdown_time(self, duckdb_cursor, create_table):
        duckdb_cursor.execute(
            """
            CREATE TABLE test_time (
                a TIME,
                b TIME,
                c TIME
            )
        """
        )
        duckdb_cursor.execute(
            """
            INSERT INTO test_time VALUES
                ('00:01:00','00:01:00','00:01:00'),
                ('00:10:00','00:10:00','00:10:00'),
                ('01:00:00','00:10:00','01:00:00'),
                (NULL,NULL,NULL)
        """
        )
        duck_tbl = duckdb_cursor.table("test_time")
        arrow_table = create_table(duck_tbl)

        # Try ==
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a ='00:01:00'").fetchone()[0] == 1
        # Try >
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a >'00:01:00'").fetchone()[0] == 2
        # Try >=
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a >='00:10:00'").fetchone()[0] == 2
        # Try <
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a <'00:10:00'").fetchone()[0] == 1
        # Try <=
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a <='00:10:00'").fetchone()[0] == 2

        # Try Is Null
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a IS NULL").fetchone()[0] == 1
        # Try Is Not Null
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a IS NOT NULL").fetchone()[0] == 3

        # Try And
        assert (
            duckdb_cursor.execute("SELECT count(*) from arrow_table where a='00:10:00' and b ='00:01:00'").fetchone()[0]
            == 0
        )
        assert (
            duckdb_cursor.execute(
                "SELECT count(*) from arrow_table where a ='01:00:00' and b = '00:10:00' and c = '01:00:00'"
            ).fetchone()[0]
            == 1
        )
        # Try Or
        assert (
            duckdb_cursor.execute("SELECT count(*) from arrow_table where a = '01:00:00' or b ='00:01:00'").fetchone()[
                0
            ]
            == 2
        )

    @pytest.mark.parametrize('create_table', [create_pyarrow_pandas, create_pyarrow_table])
    def test_filter_pushdown_timestamp(self, duckdb_cursor, create_table):
        duckdb_cursor.execute(
            """
            CREATE TABLE test_timestamp (
                a TIMESTAMP,
                b TIMESTAMP,
                c TIMESTAMP
            )
        """
        )
        duckdb_cursor.execute(
            """
            INSERT INTO test_timestamp VALUES
                ('2008-01-01 00:00:01','2008-01-01 00:00:01','2008-01-01 00:00:01'),
                ('2010-01-01 10:00:01','2010-01-01 10:00:01','2010-01-01 10:00:01'),
                ('2020-03-01 10:00:01','2010-01-01 10:00:01','2020-03-01 10:00:01'),
                (NULL,NULL,NULL)
        """
        )
        duck_tbl = duckdb_cursor.table("test_timestamp")
        arrow_table = create_table(duck_tbl)

        # Try ==
        assert (
            duckdb_cursor.execute("SELECT count(*) from arrow_table where a ='2008-01-01 00:00:01'").fetchone()[0] == 1
        )
        # Try >
        assert (
            duckdb_cursor.execute("SELECT count(*) from arrow_table where a >'2008-01-01 00:00:01'").fetchone()[0] == 2
        )
        # Try >=
        assert (
            duckdb_cursor.execute("SELECT count(*) from arrow_table where a >='2010-01-01 10:00:01'").fetchone()[0] == 2
        )
        # Try <
        assert (
            duckdb_cursor.execute("SELECT count(*) from arrow_table where a <'2010-01-01 10:00:01'").fetchone()[0] == 1
        )
        # Try <=
        assert (
            duckdb_cursor.execute("SELECT count(*) from arrow_table where a <='2010-01-01 10:00:01'").fetchone()[0] == 2
        )

        # Try Is Null
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a IS NULL").fetchone()[0] == 1
        # Try Is Not Null
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a IS NOT NULL").fetchone()[0] == 3

        # Try And
        assert (
            duckdb_cursor.execute(
                "SELECT count(*) from arrow_table where a='2010-01-01 10:00:01' and b ='2008-01-01 00:00:01'"
            ).fetchone()[0]
            == 0
        )
        assert (
            duckdb_cursor.execute(
                "SELECT count(*) from arrow_table where a ='2020-03-01 10:00:01' and b = '2010-01-01 10:00:01' and c = '2020-03-01 10:00:01'"
            ).fetchone()[0]
            == 1
        )
        # Try Or
        assert (
            duckdb_cursor.execute(
                "SELECT count(*) from arrow_table where a = '2020-03-01 10:00:01' or b ='2008-01-01 00:00:01'"
            ).fetchone()[0]
            == 2
        )

    @pytest.mark.parametrize('create_table', [create_pyarrow_pandas, create_pyarrow_table])
    def test_filter_pushdown_timestamp_TZ(self, duckdb_cursor, create_table):
        duckdb_cursor.execute(
            """
            CREATE TABLE test_timestamptz (
                a TIMESTAMPTZ,
                b TIMESTAMPTZ,
                c TIMESTAMPTZ
            )
        """
        )
        duckdb_cursor.execute(
            """
            INSERT INTO test_timestamptz VALUES
                ('2008-01-01 00:00:01','2008-01-01 00:00:01','2008-01-01 00:00:01'),
                ('2010-01-01 10:00:01','2010-01-01 10:00:01','2010-01-01 10:00:01'),
                ('2020-03-01 10:00:01','2010-01-01 10:00:01','2020-03-01 10:00:01'),
                (NULL,NULL,NULL)
            """
        )
        duck_tbl = duckdb_cursor.table("test_timestamptz")
        arrow_table = create_table(duck_tbl)

        # Try ==
        assert (
            duckdb_cursor.execute("SELECT count(*) from arrow_table where a = '2008-01-01 00:00:01'").fetchone()[0] == 1
        )
        # Try >
        assert (
            duckdb_cursor.execute("SELECT count(*) from arrow_table where a > '2008-01-01 00:00:01'").fetchone()[0] == 2
        )
        # Try >=
        assert (
            duckdb_cursor.execute("SELECT count(*) from arrow_table where a >= '2010-01-01 10:00:01'").fetchone()[0]
            == 2
        )
        # Try <
        assert (
            duckdb_cursor.execute("SELECT count(*) from arrow_table where a < '2010-01-01 10:00:01'").fetchone()[0] == 1
        )
        # Try <=
        assert (
            duckdb_cursor.execute("SELECT count(*) from arrow_table where a <= '2010-01-01 10:00:01'").fetchone()[0]
            == 2
        )

        # Try Is Null
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a IS NULL").fetchone()[0] == 1
        # Try Is Not Null
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a IS NOT NULL").fetchone()[0] == 3

        # Try And
        assert (
            duckdb_cursor.execute(
                "SELECT count(*) from arrow_table where a = '2010-01-01 10:00:01' and b = '2008-01-01 00:00:01'"
            ).fetchone()[0]
            == 0
        )
        assert (
            duckdb_cursor.execute(
                "SELECT count(*) from arrow_table where a = '2020-03-01 10:00:01' and b = '2010-01-01 10:00:01' and c = '2020-03-01 10:00:01'"
            ).fetchone()[0]
            == 1
        )
        # Try Or
        assert (
            duckdb_cursor.execute(
                "SELECT count(*) from arrow_table where a = '2020-03-01 10:00:01' or b ='2008-01-01 00:00:01'"
            ).fetchone()[0]
            == 2
        )

    @pytest.mark.parametrize('create_table', [create_pyarrow_pandas, create_pyarrow_table])
    @pytest.mark.parametrize(
        ['data_type', 'value'],
        [
            ['TINYINT', 127],
            ['SMALLINT', 32767],
            ['INTEGER', 2147483647],
            ['BIGINT', 9223372036854775807],
            ['UTINYINT', 255],
            ['USMALLINT', 65535],
            ['UINTEGER', 4294967295],
            ['UBIGINT', 18446744073709551615],
        ],
    )
    def test_filter_pushdown_integers(self, duckdb_cursor, data_type, value, create_table):
        duckdb_cursor.execute(
            f"""
            CREATE TABLE tbl as select {value}::{data_type} as i
        """
        )
        expected = duckdb_cursor.table('tbl').fetchall()
        filter = "i > 0"
        rel = duckdb_cursor.table('tbl')
        arrow_table = create_table(rel)
        actual = duckdb_cursor.sql(f"select * from arrow_table where {filter}").fetchall()
        assert expected == actual

        # Test with equivalent prepared statement
        actual = duckdb_cursor.execute("select * from arrow_table where i > ?", (0,)).fetchall()
        assert expected == actual
        # Test equality
        actual = duckdb_cursor.execute("select * from arrow_table where i = ?", (value,)).fetchall()
        assert expected == actual

    @pytest.mark.skipif(
        Version(pa.__version__) < Version('15.0.0'), reason="pyarrow 14.0.2 'to_pandas' causes a DeprecationWarning"
    )
    def test_9371(self, duckdb_cursor, tmp_path):
        import datetime
        import pathlib

        # connect to an in-memory database
        duckdb_cursor.execute("SET TimeZone='UTC';")
        base_path = tmp_path / "parquet_folder"
        base_path.mkdir(exist_ok=True)
        file_path = base_path / "test.parquet"

        duckdb_cursor.execute("SET TimeZone='UTC';")

        # Example data
        dt = datetime.datetime(2023, 8, 29, 1, tzinfo=datetime.timezone.utc)

        my_arrow_table = pa.Table.from_pydict({'ts': [dt, dt, dt], 'value': [1, 2, 3]})
        df = my_arrow_table.to_pandas()
        df = df.set_index("ts")  # SET INDEX! (It all works correctly when the index is not set)
        df.to_parquet(str(file_path))

        my_arrow_dataset = ds.dataset(str(file_path))
        res = duckdb_cursor.execute("SELECT * FROM my_arrow_dataset WHERE ts = ?", parameters=[dt]).arrow()
        output = duckdb_cursor.sql("select * from res").fetchall()
        expected = [(1, dt), (2, dt), (3, dt)]
        assert output == expected

    @pytest.mark.parametrize('create_table', [create_pyarrow_pandas, create_pyarrow_table])
    def test_filter_pushdown_date(self, duckdb_cursor, create_table):
        duckdb_cursor.execute(
            """
            CREATE TABLE test_date (
                a DATE,
                b DATE,
                c DATE
            )
        """
        )
        duckdb_cursor.execute(
            """
            INSERT INTO test_date VALUES
                ('2000-01-01','2000-01-01','2000-01-01'),
                ('2000-10-01','2000-10-01','2000-10-01'),
                ('2010-01-01','2000-10-01','2010-01-01'),
                (NULL,NULL,NULL)
        """
        )
        duck_tbl = duckdb_cursor.table("test_date")
        arrow_table = create_table(duck_tbl)

        # Try ==
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a = '2000-01-01'").fetchone()[0] == 1
        # Try >
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a > '2000-01-01'").fetchone()[0] == 2
        # Try >=
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a >= '2000-10-01'").fetchone()[0] == 2
        # Try <
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a < '2000-10-01'").fetchone()[0] == 1
        # Try <=
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a <= '2000-10-01'").fetchone()[0] == 2

        # Try Is Null
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a IS NULL").fetchone()[0] == 1
        # Try Is Not Null
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a IS NOT NULL").fetchone()[0] == 3

        # Try And
        assert (
            duckdb_cursor.execute(
                "SELECT count(*) from arrow_table where a = '2000-10-01' and b = '2000-01-01'"
            ).fetchone()[0]
            == 0
        )
        assert (
            duckdb_cursor.execute(
                "SELECT count(*) from arrow_table where a = '2010-01-01' and b = '2000-10-01' and c = '2010-01-01'"
            ).fetchone()[0]
            == 1
        )
        # Try Or
        assert (
            duckdb_cursor.execute(
                "SELECT count(*) from arrow_table where a = '2010-01-01' or b = '2000-01-01'"
            ).fetchone()[0]
            == 2
        )

    @pytest.mark.parametrize('create_table', [create_pyarrow_pandas, create_pyarrow_table])
    def test_filter_pushdown_blob(self, duckdb_cursor, create_table):
        import pandas

        df = pandas.DataFrame(
            {
                'a': [bytes([1]), bytes([2]), bytes([3]), None],
                'b': [bytes([1]), bytes([2]), bytes([3]), None],
                'c': [bytes([1]), bytes([2]), bytes([3]), None],
            }
        )
        rel = duckdb.from_df(df)
        arrow_table = create_table(rel)

        # Try ==
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a = '\x01'").fetchone()[0] == 1
        # # Try >
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a > '\x01'").fetchone()[0] == 2
        # Try >=
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a >= '\x02'").fetchone()[0] == 2
        # Try <
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a < '\x02'").fetchone()[0] == 1
        # Try <=
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a <= '\x02'").fetchone()[0] == 2

        # Try Is Null
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a IS NULL").fetchone()[0] == 1
        # Try Is Not Null
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a IS NOT NULL").fetchone()[0] == 3

        # Try And
        assert duckdb_cursor.execute("SELECT count(*) from arrow_table where a='\x02' and b ='\x01'").fetchone()[0] == 0
        assert (
            duckdb_cursor.execute(
                "SELECT count(*) from arrow_table where a = '\x02' and b = '\x02' and c = '\x02'"
            ).fetchone()[0]
            == 1
        )
        # Try Or
        assert (
            duckdb_cursor.execute("SELECT count(*) from arrow_table where a = '\x01' or b = '\x02'").fetchone()[0] == 2
        )

    @pytest.mark.parametrize('create_table', [create_pyarrow_pandas, create_pyarrow_table, create_pyarrow_dataset])
    def test_filter_pushdown_no_projection(self, duckdb_cursor, create_table):
        duckdb_cursor.execute(
            """
            CREATE TABLE test_int (
                a INTEGER,
                b INTEGER,
                c INTEGER
            )
        """
        )
        duckdb_cursor.execute(
            """
            INSERT INTO test_int VALUES
                (1,1,1),
                (10,10,10),
                (100,10,100),
                (NULL,NULL,NULL)
        """
        )
        duck_tbl = duckdb_cursor.table("test_int")
        arrow_table = create_table(duck_tbl)

        assert duckdb_cursor.execute("SELECT * FROM arrow_table VALUES where a = 1").fetchall() == [(1, 1, 1)]

    @pytest.mark.parametrize('create_table', [create_pyarrow_pandas, create_pyarrow_table])
    def test_filter_pushdown_2145(self, duckdb_cursor, tmp_path, create_table):
        import pandas

        date1 = pandas.date_range("2018-01-01", "2018-12-31", freq="B")
        df1 = pandas.DataFrame(np.random.randn(date1.shape[0], 5), columns=list("ABCDE"))
        df1["date"] = date1

        date2 = pandas.date_range("2019-01-01", "2019-12-31", freq="B")
        df2 = pandas.DataFrame(np.random.randn(date2.shape[0], 5), columns=list("ABCDE"))
        df2["date"] = date2

        data1 = tmp_path / 'data1.parquet'
        data2 = tmp_path / 'data2.parquet'
        duckdb_cursor.execute(f"copy (select * from df1) to '{data1.as_posix()}'")
        duckdb_cursor.execute(f"copy (select * from df2) to '{data2.as_posix()}'")

        glob_pattern = tmp_path / 'data*.parquet'
        table = duckdb_cursor.read_parquet(glob_pattern.as_posix()).arrow()

        output_df = duckdb.arrow(table).filter("date > '2019-01-01'").df()
        expected_df = duckdb.from_parquet(glob_pattern.as_posix()).filter("date > '2019-01-01'").df()
        pandas.testing.assert_frame_equal(expected_df, output_df)

    # https://github.com/duckdb/duckdb/pull/4817/files#r1339973721
    @pytest.mark.parametrize('create_table', [create_pyarrow_pandas, create_pyarrow_table])
    def test_filter_column_removal(self, duckdb_cursor, create_table):
        duckdb_cursor.execute(
            """
            CREATE TABLE test AS SELECT
                range a,
                100 - range b
            FROM range(100)
        """
        )
        duck_test_table = duckdb_cursor.table("test")
        arrow_table = create_table(duck_test_table)

        # PR 4817 - remove filter columns that are unused in the remainder of the query plan from the table function
        query_res = duckdb_cursor.execute(
            """
            EXPLAIN SELECT count(*) FROM arrow_table WHERE
                a > 25 AND b > 25
        """
        ).fetchall()

        # scanned columns that come out of the scan are displayed like this, so we shouldn't see them
        match = re.search("│ +a +│", query_res[0][1])
        assert not match
        match = re.search("│ +b +│", query_res[0][1])
        assert not match

    @pytest.mark.skipif(sys.version_info < (3, 9), reason="Requires python 3.9")
    @pytest.mark.parametrize('create_table', [create_pyarrow_pandas, create_pyarrow_table])
    def test_struct_filter_pushdown(self, duckdb_cursor, create_table):
        duckdb_cursor.execute(
            """
            CREATE TABLE test_structs (s STRUCT(a integer, b bool))
        """
        )
        duckdb_cursor.execute(
            """
            INSERT INTO test_structs VALUES
                ({'a': 1, 'b': true}), 
                ({'a': 2, 'b': false}), 
                (NULL),
                ({'a': 3, 'b': true}), 
                ({'a': NULL, 'b': NULL});
        """
        )

        duck_tbl = duckdb_cursor.table("test_structs")
        arrow_table = create_table(duck_tbl)

        # Ensure that the filter is pushed down
        query_res = duckdb_cursor.execute(
            """
            EXPLAIN SELECT * FROM arrow_table WHERE
                s.a < 2
        """
        ).fetchall()

        input = query_res[0][1]
        if 'PANDAS_SCAN' in input:
            pytest.skip(reason="This version of pandas does not produce an Arrow object")
        match = re.search(".*ARROW_SCAN.*Filters: s\\.a<2 AND s\\.a IS.*NOT NULL.*", input, flags=re.DOTALL)
        assert match

        # Check that the filter is applied correctly
        assert duckdb_cursor.execute("SELECT * FROM arrow_table WHERE s.a < 2").fetchone()[0] == {"a": 1, "b": True}

        query_res = duckdb_cursor.execute(
            """
            EXPLAIN SELECT * FROM arrow_table WHERE s.a < 3 AND s.b = true 
        """
        ).fetchall()

        # the explain-output is pretty cramped, so just make sure we see both struct references.
        match = re.search(
            ".*ARROW_SCAN.*Filters: s\\.a<3 AND s\\.a IS.*NOT NULL AND s\\.b=true\\.\\.\\..*\\.b IS NOT NULL.*",
            query_res[0][1],
            flags=re.DOTALL,
        )
        assert match

        # Check that the filter is applied correctly
        assert duckdb_cursor.execute("SELECT COUNT(*) FROM arrow_table WHERE s.a < 3 AND s.b = true").fetchone()[0] == 1
        assert duckdb_cursor.execute("SELECT * FROM arrow_table WHERE s.a < 3 AND s.b = true").fetchone()[0] == {
            "a": 1,
            "b": True,
        }

        # This should not produce a pushdown
        query_res = duckdb_cursor.execute(
            """
            EXPLAIN SELECT * FROM arrow_table WHERE
                s.a IS NULL
        """
        ).fetchall()

        match = re.search(".*ARROW_SCAN.*Filters: s\\.a IS NULL.*", query_res[0][1], flags=re.DOTALL)
        assert not match

    @pytest.mark.skipif(sys.version_info < (3, 9), reason="Requires python 3.9")
    @pytest.mark.parametrize('create_table', [create_pyarrow_pandas, create_pyarrow_table])
    def test_nested_struct_filter_pushdown(self, duckdb_cursor, create_table):
        duckdb_cursor.execute(
            """
            CREATE TABLE test_nested_structs(s STRUCT(a STRUCT(b integer, c bool), d STRUCT(e integer, f varchar)));
        """
        )
        duckdb_cursor.execute(
            """
            INSERT INTO test_nested_structs VALUES
                ({'a': {'b': 1, 'c': false}, 'd': {'e': 2, 'f': 'foo'}}),
                (NULL),
                ({'a': {'b': 3, 'c': true}, 'd': {'e': 4, 'f': 'bar'}}),
                ({'a': {'b': NULL, 'c': true}, 'd': {'e': 5, 'f': 'qux'}}),
                ({'a': NULL, 'd': NULL});
        """
        )

        duck_tbl = duckdb_cursor.table("test_nested_structs")
        arrow_table = create_table(duck_tbl)

        # Ensure that the filter is pushed down
        query_res = duckdb_cursor.execute(
            """
            EXPLAIN SELECT * FROM arrow_table WHERE s.a.b < 2;
        """
        ).fetchall()

        input = query_res[0][1]
        if 'PANDAS_SCAN' in input:
            pytest.skip(reason="This version of pandas does not produce an Arrow object")
        match = re.search(".*ARROW_SCAN.*Filters: s\\.a\\.b<2 AND s\\.a\\.b.*IS NOT NULL.*", input, flags=re.DOTALL)
        assert match

        # Check that the filter is applied correctly
        assert duckdb_cursor.execute("SELECT * FROM arrow_table WHERE s.a.b < 2").fetchone()[0] == {
            'a': {'b': 1, 'c': False},
            'd': {'e': 2, 'f': 'foo'},
        }

        query_res = duckdb_cursor.execute(
            """
            EXPLAIN SELECT * FROM arrow_table WHERE s.a.c=true AND s.d.e=5 
        """
        ).fetchall()

        # the explain-output is pretty cramped, so just make sure we see both struct references.
        match = re.search(
            ".*ARROW_SCAN.*Filters: s\\.a\\.c=true AND s\\.a.*\\.c IS NOT NULL.*AND s\\.d\\.e=5.*AND s\\.d\\.e.*IS NOT NULL.*",
            query_res[0][1],
            flags=re.DOTALL,
        )
        assert match

        # Check that the filter is applied correctly
        assert duckdb_cursor.execute("SELECT COUNT(*) FROM arrow_table WHERE s.a.c=true AND s.d.e=5").fetchone()[0] == 1
        assert duckdb_cursor.execute("SELECT * FROM arrow_table WHERE s.a.c=true AND s.d.e=5").fetchone()[0] == {
            'a': {'b': None, 'c': True},
            'd': {'e': 5, 'f': 'qux'},
        }

        query_res = duckdb_cursor.execute(
            """
            EXPLAIN SELECT * FROM arrow_table WHERE s.d.f = 'bar';
        """
        )

        match = re.search(
            ".*ARROW_SCAN.*Filters: s\\.d\\.f=bar AND s\\.d.*\\.f IS NOT NULL.*",
            query_res.fetchone()[1],
            flags=re.DOTALL,
        )

        assert match

        # Check that the filter is applied correctly
        assert duckdb_cursor.execute("SELECT * FROM arrow_table WHERE s.d.f = 'bar'").fetchone()[0] == {
            'a': {'b': 3, 'c': True},
            'd': {'e': 4, 'f': 'bar'},
        }



import duckdb
import pytest
from conftest import NumpyPandas, ArrowPandas


class TestAppendDF(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_df_to_table_append(self, duckdb_cursor, pandas):
        conn = duckdb.connect()
        conn.execute("Create table integers (i integer)")
        df_in = pandas.DataFrame(
            {
                'numbers': [1, 2, 3, 4, 5],
            }
        )
        conn.append('integers', df_in)
        assert conn.execute('select count(*) from integers').fetchone()[0] == 5

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_append_by_name(self, pandas):
        con = duckdb.connect()
        con.execute("create table tbl (a integer, b bool, c varchar)")
        df_in = pandas.DataFrame({'c': ['duck', 'db'], 'b': [False, True], 'a': [4, 2]})
        # By default we append by position, causing the following exception:
        with pytest.raises(
            duckdb.ConversionException, match="Conversion Error: Could not convert string 'duck' to INT32"
        ):
            con.append('tbl', df_in)

        # When we use 'by_name' we instead append by name
        con.append('tbl', df_in, by_name=True)
        res = con.table('tbl').fetchall()
        assert res == [(4, False, 'duck'), (2, True, 'db')]

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_append_by_name_quoted(self, pandas):
        con = duckdb.connect()
        con.execute(
            """
            create table tbl ("needs to be quoted" integer, other varchar)
        """
        )
        df_in = pandas.DataFrame({"needs to be quoted": [1, 2, 3]})
        con.append('tbl', df_in, by_name=True)
        res = con.table('tbl').fetchall()
        assert res == [(1, None), (2, None), (3, None)]

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_append_by_name_no_exact_match(self, pandas):
        con = duckdb.connect()
        con.execute("create table tbl (a integer, b bool)")
        df_in = pandas.DataFrame({'c': ['a', 'b'], 'b': [True, False], 'a': [42, 1337]})
        # Too many columns raises an error, because the columns cant be found in the targeted table
        with pytest.raises(duckdb.BinderException, match='Table "tbl" does not have a column with name "c"'):
            con.append('tbl', df_in, by_name=True)

        df_in = pandas.DataFrame({'b': [False, False, False]})

        # Not matching all columns is not a problem, as they will be filled with NULL instead
        con.append('tbl', df_in, by_name=True)
        res = con.table('tbl').fetchall()
        # 'a' got filled by NULL automatically because it wasn't inserted into
        assert res == [(None, False), (None, False), (None, False)]

        # Empty the table
        con.execute("create or replace table tbl (a integer, b bool)")

        df_in = pandas.DataFrame({'a': [1, 2, 3]})
        con.append('tbl', df_in, by_name=True)
        res = con.table('tbl').fetchall()
        # Also works for missing columns *after* the supplied ones
        assert res == [(1, None), (2, None), (3, None)]



import duckdb
import tempfile
import os
import pytest
from conftest import NumpyPandas, ArrowPandas


class TestInsert(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_insert(self, pandas):
        test_df = pandas.DataFrame({"i": [1, 2, 3], "j": ["one", "two", "three"]})
        # connect to an in-memory temporary database
        conn = duckdb.connect()
        # get a cursor
        cursor = conn.cursor()
        conn.execute("CREATE TABLE test (i INTEGER, j STRING)")
        rel = conn.table("test")
        rel.insert([1, 'one'])
        rel.insert([2, 'two'])
        rel.insert([3, 'three'])
        rel_a3 = cursor.table('test').project('CAST(i as BIGINT)i, j').to_df()
        pandas.testing.assert_frame_equal(rel_a3, test_df)



import duckdb
import pytest


class TestJoin(object):
    def test_alias_from_sql(self):
        con = duckdb.connect()
        rel1 = con.sql("SELECT 1 AS col1, 2 AS col2")
        rel2 = con.sql("SELECT 1 AS col1, 3 AS col3")

        rel = con.sql('select * from rel1 JOIN rel2 USING (col1)')
        rel.show()
        res = rel.fetchall()
        assert res == [(1, 2, 3)]

    def test_relational_join(self):
        con = duckdb.connect()

        rel1 = con.sql("SELECT 1 AS col1, 2 AS col2")
        rel2 = con.sql("SELECT 1 AS col1, 3 AS col3")

        rel = rel1.join(rel2, 'col1')
        res = rel.fetchall()
        assert res == [(1, 2, 3)]

    def test_relational_join_alias_collision(self):
        con = duckdb.connect()

        rel1 = con.sql("SELECT 1 AS col1, 2 AS col2").set_alias('a')
        rel2 = con.sql("SELECT 1 AS col1, 3 AS col3").set_alias('a')

        with pytest.raises(duckdb.InvalidInputException, match='Both relations have the same alias'):
            rel = rel1.join(rel2, 'col1')

    def test_relational_join_with_condition(self):
        con = duckdb.connect()

        rel1 = con.sql("SELECT 1 AS col1, 2 AS col2", alias='rel1')
        rel2 = con.sql("SELECT 1 AS col1, 3 AS col3", alias='rel2')

        # This makes a USING clause, which is kind of unexpected behavior
        rel = rel1.join(rel2, 'rel1.col1 = rel2.col1')
        rel.show()
        res = rel.fetchall()
        assert res == [(1, 2, 1, 3)]



import duckdb
import pytest
import sys
import datetime
import os

adbc_driver_manager = pytest.importorskip("adbc_driver_manager.dbapi")
adbc_driver_manager_lib = pytest.importorskip("adbc_driver_manager._lib")

pyarrow = pytest.importorskip("pyarrow")

# When testing local, if you build via BUILD_PYTHON=1 make, you need to manually set up the
# dylib duckdb path.
driver_path = duckdb.duckdb.__file__


@pytest.fixture
def duck_conn():
    with adbc_driver_manager.connect(driver=driver_path, entrypoint="duckdb_adbc_init") as conn:
        yield conn


def example_table():
    return pyarrow.table(
        [
            [1, 2, 3, 4],
            ["a", "b", None, "d"],
        ],
        names=["ints", "strs"],
    )


@pytest.mark.xfail
def test_connection_get_info(duck_conn):
    assert duck_conn.adbc_get_info() != {}


def test_connection_get_table_types(duck_conn):
    assert duck_conn.adbc_get_table_types() == []
    with duck_conn.cursor() as cursor:
        # Test Default Schema
        cursor.execute("CREATE TABLE tableschema (ints BIGINT)")
    assert duck_conn.adbc_get_table_types() == ['BASE TABLE']


def test_connection_get_objects(duck_conn):
    with duck_conn.cursor() as cursor:
        cursor.execute("CREATE TABLE getobjects (ints BIGINT)")
    assert duck_conn.adbc_get_objects(depth="all").read_all().to_pylist() == [
        {
            'db_schema_name': 'main',
            'db_schema_tables': [
                {
                    'table_name': 'getobjects',
                    'table_columns': [{'column_name': 'ints', 'ordinal_position': 2, 'remarks': ''}],
                }
            ],
        }
    ]


def test_commit(tmp_path):
    db = os.path.join(tmp_path, "tmp.db")
    if os.path.exists(db):
        os.remove(db)
    table = example_table()
    db_kwargs = {"path": f"{db}"}
    # Start connection with auto-commit off
    with adbc_driver_manager.connect(
        driver=driver_path,
        entrypoint="duckdb_adbc_init",
        db_kwargs=db_kwargs,
    ) as conn:
        assert not conn._autocommit
        with conn.cursor() as cur:
            cur.adbc_ingest("ingest", table, "create")

    # Check Data is not there
    with adbc_driver_manager.connect(
        driver=driver_path,
        entrypoint="duckdb_adbc_init",
        db_kwargs=db_kwargs,
        autocommit=True,
    ) as conn:
        assert conn._autocommit
        with conn.cursor() as cur:
            # This errors because the table does not exist
            with pytest.raises(
                adbc_driver_manager_lib.InternalError,
                match=r'Table with name ingest does not exist!',
            ):
                cur.execute("SELECT count(*) from ingest")

            cur.adbc_ingest("ingest", table, "create")

    # This now works because we enabled autocommit
    with adbc_driver_manager.connect(
        driver=driver_path,
        entrypoint="duckdb_adbc_init",
        db_kwargs=db_kwargs,
    ) as conn:
        with conn.cursor() as cur:
            cur.execute("SELECT count(*) from ingest")
            assert cur.fetch_arrow_table().to_pydict() == {'count_star()': [4]}


def test_connection_get_table_schema(duck_conn):
    with duck_conn.cursor() as cursor:
        # Test Default Schema
        cursor.execute("CREATE TABLE tableschema (ints BIGINT)")
        assert duck_conn.adbc_get_table_schema("tableschema") == pyarrow.schema(
            [
                ("ints", "int64"),
            ]
        )

        # Test Given Schema
        cursor.execute("CREATE SCHEMA test;")
        cursor.execute("CREATE TABLE test.tableschema (test_ints BIGINT)")
        assert duck_conn.adbc_get_table_schema("tableschema", db_schema_filter="test") == pyarrow.schema(
            [
                ("test_ints", "int64"),
            ]
        )
        assert duck_conn.adbc_get_table_schema("tableschema") == pyarrow.schema(
            [
                ("ints", "int64"),
            ]
        )

        # Catalog name is currently not supported
        with pytest.raises(
            adbc_driver_manager_lib.NotSupportedError,
            match=r'Catalog Name is not used in DuckDB. It must be set to nullptr or an empty string',
        ):
            duck_conn.adbc_get_table_schema("tableschema", catalog_filter="bla", db_schema_filter="test")


def test_prepared_statement(duck_conn):
    with duck_conn.cursor() as cursor:
        cursor.adbc_prepare("SELECT 1")
        cursor.execute("SELECT 1")
        assert cursor.fetchone() == (1,)
        assert cursor.fetchone() is None


def test_statement_query(duck_conn):
    with duck_conn.cursor() as cursor:
        cursor.execute("SELECT 1")
        assert cursor.fetchone() == (1,)
        assert cursor.fetchone() is None

        cursor.execute("SELECT 1 AS foo")
        assert cursor.fetch_arrow_table().to_pylist() == [{"foo": 1}]


def test_insertion(duck_conn):
    table = example_table()
    reader = table.to_reader()

    with duck_conn.cursor() as cursor:
        cursor.adbc_ingest("ingest", reader, "create")
        cursor.execute("SELECT * FROM ingest")
        assert cursor.fetch_arrow_table() == table

    with duck_conn.cursor() as cursor:
        cursor.adbc_ingest("ingest_table", table, "create")
        cursor.execute("SELECT * FROM ingest")
        assert cursor.fetch_arrow_table() == table

    # Test Append
    with duck_conn.cursor() as cursor:
        with pytest.raises(
            adbc_driver_manager_lib.InternalError,
            match=r'Failed to create table \'ingest_table\': Table with name "ingest_table" already exists!',
        ):
            cursor.adbc_ingest("ingest_table", table, "create")
        cursor.adbc_ingest("ingest_table", table, "append")
        cursor.execute("SELECT count(*) FROM ingest_table")
        assert cursor.fetch_arrow_table().to_pydict() == {"count_star()": [8]}


def test_read(duck_conn):
    with duck_conn.cursor() as cursor:
        filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "data", "category.csv")
        cursor.execute(f"SELECT * FROM '{filename}'")
        assert cursor.fetch_arrow_table().to_pydict() == {
            "CATEGORY_ID": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16],
            "NAME": [
                "Action",
                "Animation",
                "Children",
                "Classics",
                "Comedy",
                "Documentary",
                "Drama",
                "Family",
                "Foreign",
                "Games",
                "Horror",
                "Music",
                "New",
                "Sci-Fi",
                "Sports",
                "Travel",
            ],
            "LAST_UPDATE": [
                datetime.datetime(2006, 2, 15, 4, 46, 27),
                datetime.datetime(2006, 2, 15, 4, 46, 27),
                datetime.datetime(2006, 2, 15, 4, 46, 27),
                datetime.datetime(2006, 2, 15, 4, 46, 27),
                datetime.datetime(2006, 2, 15, 4, 46, 27),
                datetime.datetime(2006, 2, 15, 4, 46, 27),
                datetime.datetime(2006, 2, 15, 4, 46, 27),
                datetime.datetime(2006, 2, 15, 4, 46, 27),
                datetime.datetime(2006, 2, 15, 4, 46, 27),
                datetime.datetime(2006, 2, 15, 4, 46, 27),
                datetime.datetime(2006, 2, 15, 4, 46, 27),
                datetime.datetime(2006, 2, 15, 4, 46, 27),
                datetime.datetime(2006, 2, 15, 4, 46, 27),
                datetime.datetime(2006, 2, 15, 4, 46, 27),
                datetime.datetime(2006, 2, 15, 4, 46, 27),
                datetime.datetime(2006, 2, 15, 4, 46, 27),
            ],
        }



import pytest
import duckdb
import pandas as pd
import numpy as np


def compare_results(query, list_values=[], list_values_2=[]):
    con = duckdb.connect()
    df_duck = con.query(query).df()
    duck_values = df_duck['a']
    for duck_value in duck_values:
        is_in_list_value = False
        for value in list_values:
            if duck_value == value:
                is_in_list_value = True
        for value in list_values_2:
            if duck_value == value:
                is_in_list_value = True
        assert is_in_list_value


class TestFetchNested(object):
    @pytest.mark.skip(reason="sporadically breaks in CI")
    def test_fetch_df_list(self):
        con = duckdb.connect()
        # Integers
        compare_results("SELECT a from (select list_value(3,5,10) as a) as t", [[3, 5, 10]])
        compare_results("SELECT a from (select list_value(3,5,NULL) as a) as t", [[3, 5, None]])
        compare_results("SELECT a from (select list_value(NULL,NULL,NULL) as a) as t", [[None, None, None]])
        compare_results("SELECT a from (select list_value() as a) as t", [[]])

        # Strings
        compare_results(
            "SELECT a from (select list_value('test','test_one','test_two') as a) as t",
            [['test', 'test_one', 'test_two']],
        )
        compare_results(
            "SELECT a from (select list_value('test','test_one',NULL) as a) as t", [['test', 'test_one', None]]
        )

        # Big Lists
        compare_results("SELECT a from (SELECT LIST(i) as a FROM range(10000) tbl(i)) as t", [list(range(0, 10000))])

        # #Multiple Lists
        compare_results(
            "SELECT a from (SELECT LIST(i) as a FROM range(5) tbl(i) group by i%2 order by all) as t",
            [[0, 2, 4], [1, 3]],
        )

        # Unique Constants
        compare_results(
            "SELECT a from (SELECT list_value(1) as a FROM range(5) tbl(i)) as t", [[1], [1], [1], [1], [1]]
        )

        # Nested Lists
        compare_results(
            "SELECT LIST(le) as a FROM (SELECT LIST(i) le from range(5) tbl(i) group by i%2 order by all) as t order by all",
            [[[0, 2, 4], [1, 3]]],
            [[[1, 3], [0, 2, 4]]],
        )

        # LIST[LIST[LIST[LIST[LIST[INTEGER]]]]]]
        compare_results(
            "SELECT list (lllle)  as a from (SELECT list (llle) lllle from (SELECT list(lle) llle from (SELECT LIST(le) lle FROM (SELECT LIST(i) le from range(5) tbl(i) group by i%2 order by all) as t order by all) as t1 order by all) as t2 order by all) as t3 order by all",
            [[[[[[0, 2, 4], [1, 3]]]]]],
            [[[[[[1, 3], [0, 2, 4]]]]]],
        )

        compare_results(
            '''SELECT grp,lst,a FROM (select grp, lst, case when grp>1 then lst else list_value(null) end as a
                         from (SELECT a_1%4 as grp, list(a_1) as lst FROM range(7) tbl(a_1) group by grp order by all) as lst_tbl) as T;''',
            [[None], [None], [2, 6], [3]],
        )

        # Tests for converting multiple lists to/from Pandas with NULL values and/or strings
        compare_results(
            "SELECT list(st) as a from (select i, case when i%5 then NULL else i::VARCHAR end as st from range(10) tbl(i)) as t group by i%2 order by all",
            [['0', None, None, None, None], [None, None, '5', None, None]],
        )

    @pytest.mark.skip(reason="sporadically breaks in CI")
    def test_struct_df(self):
        compare_results("SELECT a from (SELECT STRUCT_PACK(a := 42, b := 43) as a) as t", [{'a': 42, 'b': 43}])

        compare_results("SELECT a from (SELECT STRUCT_PACK(a := NULL, b := 43) as a) as t", [{'a': None, 'b': 43}])

        compare_results("SELECT a from (SELECT STRUCT_PACK(a := NULL) as a) as t", [{'a': None}])

        compare_results(
            "SELECT a from (SELECT STRUCT_PACK(a := i, b := i) as a FROM range(10) tbl(i)) as t",
            [
                {'a': 0, 'b': 0},
                {'a': 1, 'b': 1},
                {'a': 2, 'b': 2},
                {'a': 3, 'b': 3},
                {'a': 4, 'b': 4},
                {'a': 5, 'b': 5},
                {'a': 6, 'b': 6},
                {'a': 7, 'b': 7},
                {'a': 8, 'b': 8},
                {'a': 9, 'b': 9},
            ],
        )

        compare_results(
            "SELECT a from (SELECT STRUCT_PACK(a := LIST_VALUE(1,2,3), b := i) as a FROM range(10) tbl(i)) as t",
            [
                {'a': [1, 2, 3], 'b': 0},
                {'a': [1, 2, 3], 'b': 1},
                {'a': [1, 2, 3], 'b': 2},
                {'a': [1, 2, 3], 'b': 3},
                {'a': [1, 2, 3], 'b': 4},
                {'a': [1, 2, 3], 'b': 5},
                {'a': [1, 2, 3], 'b': 6},
                {'a': [1, 2, 3], 'b': 7},
                {'a': [1, 2, 3], 'b': 8},
                {'a': [1, 2, 3], 'b': 9},
            ],
        )

    @pytest.mark.skip(reason="sporadically breaks in CI")
    def test_map_df(self):
        compare_results(
            "SELECT a from (select MAP(LIST_VALUE(1, 2, 3, 4),LIST_VALUE(10, 9, 8, 7)) as a) as t",
            [{'key': [1, 2, 3, 4], 'value': [10, 9, 8, 7]}],
        )

        with pytest.raises(duckdb.InvalidInputException, match="Map keys have to be unique"):
            compare_results(
                "SELECT a from (select MAP(LIST_VALUE(1, 2, 3, 4,2, NULL),LIST_VALUE(10, 9, 8, 7,11,42)) as a) as t",
                [{'key': [1, 2, 3, 4, 2, None], 'value': [10, 9, 8, 7, 11, 42]}],
            )

        compare_results("SELECT a from (select MAP(LIST_VALUE(),LIST_VALUE()) as a) as t", [{'key': [], 'value': []}])

        with pytest.raises(duckdb.InvalidInputException, match="Map keys have to be unique"):
            compare_results(
                "SELECT a from (select MAP(LIST_VALUE('Jon Lajoie', 'Backstreet Boys', 'Tenacious D','Jon Lajoie' ),LIST_VALUE(10,9,10,11)) as a) as t",
                [{'key': ['Jon Lajoie', 'Backstreet Boys', 'Tenacious D', 'Jon Lajoie'], 'value': [10, 9, 10, 11]}],
            )

        with pytest.raises(duckdb.InvalidInputException, match="Map keys can not be NULL"):
            compare_results(
                "SELECT a from (select MAP(LIST_VALUE('Jon Lajoie', NULL, 'Tenacious D',NULL,NULL ),LIST_VALUE(10,9,10,11,13)) as a) as t",
                [{'key': ['Jon Lajoie', None, 'Tenacious D', None, None], 'value': [10, 9, 10, 11, 13]}],
            )

        with pytest.raises(duckdb.InvalidInputException, match="Map keys can not be NULL"):
            compare_results(
                "SELECT a from (select MAP(LIST_VALUE(NULL, NULL, NULL,NULL,NULL ),LIST_VALUE(10,9,10,11,13)) as a) as t",
                [{'key': [None, None, None, None, None], 'value': [10, 9, 10, 11, 13]}],
            )

        with pytest.raises(duckdb.InvalidInputException, match="Map keys can not be NULL"):
            compare_results(
                "SELECT a from (select MAP(LIST_VALUE(NULL, NULL, NULL,NULL,NULL ),LIST_VALUE(NULL, NULL, NULL,NULL,NULL )) as a) as t",
                [{'key': [None, None, None, None, None], 'value': [None, None, None, None, None]}],
            )

        compare_results(
            "SELECT m as a from (select MAP(list_value(1), list_value(2)) from range(5) tbl(i)) tbl(m)",
            [
                {'key': [1], 'value': [2]},
                {'key': [1], 'value': [2]},
                {'key': [1], 'value': [2]},
                {'key': [1], 'value': [2]},
                {'key': [1], 'value': [2]},
            ],
        )

        compare_results(
            "SELECT m as a from (select MAP(lsta,lstb) as m from (SELECT list(i) as lsta, list(i) as lstb from range(10) tbl(i) group by i%5 order by all) as lst_tbl) as T",
            [
                {'key': [0, 5], 'value': [0, 5]},
                {'key': [1, 6], 'value': [1, 6]},
                {'key': [2, 7], 'value': [2, 7]},
                {'key': [3, 8], 'value': [3, 8]},
                {'key': [4, 9], 'value': [4, 9]},
            ],
        )

    @pytest.mark.skip(reason="sporadically breaks in CI")
    def test_nested_mix(self):
        con = duckdb.connect()
        # List of structs W/ Struct that is NULL entirely
        compare_results(
            "SELECT [{'i':1,'j':2},NULL,{'i':2,'j':NULL}] as a", [[{'i': 1, 'j': 2}, None, {'i': 2, 'j': None}]]
        )

        # Lists of structs with lists
        compare_results("SELECT [{'i':1,'j':[2,3]},NULL] as a", [[{'i': 1, 'j': [2, 3]}, None]])

        # Maps embedded in a struct
        compare_results(
            "SELECT {'i':mp,'j':mp2} as a FROM (SELECT MAP(LIST_VALUE(1, 2, 3, 4),LIST_VALUE(10, 9, 8, 7)) as mp, MAP(LIST_VALUE(1, 2, 3, 5),LIST_VALUE(10, 9, 8, 7)) as mp2) as t",
            [{'i': {'key': [1, 2, 3, 4], 'value': [10, 9, 8, 7]}, 'j': {'key': [1, 2, 3, 5], 'value': [10, 9, 8, 7]}}],
        )

        # List of maps
        compare_results(
            "SELECT [mp,mp2] as a FROM (SELECT MAP(LIST_VALUE(1, 2, 3, 4),LIST_VALUE(10, 9, 8, 7)) as mp, MAP(LIST_VALUE(1, 2, 3, 5),LIST_VALUE(10, 9, 8, 7)) as mp2) as t",
            [[{'key': [1, 2, 3, 4], 'value': [10, 9, 8, 7]}, {'key': [1, 2, 3, 5], 'value': [10, 9, 8, 7]}]],
        )

        # Map with list as key and/or value
        compare_results(
            "SELECT MAP(LIST_VALUE([1,2],[3,4],[5,4]),LIST_VALUE([1,2],[3,4],[5,4])) as a",
            [{'key': [[1, 2], [3, 4], [5, 4]], 'value': [[1, 2], [3, 4], [5, 4]]}],
        )

        # Map with struct as key and/or value
        compare_results(
            "SELECT MAP(LIST_VALUE({'i':1,'j':2},{'i':3,'j':4}),LIST_VALUE({'i':1,'j':2},{'i':3,'j':4})) as a",
            [{'key': [{'i': 1, 'j': 2}, {'i': 3, 'j': 4}], 'value': [{'i': 1, 'j': 2}, {'i': 3, 'j': 4}]}],
        )

        # Null checks on lists with structs
        compare_results(
            "SELECT [{'i':1,'j':[2,3]},NULL,{'i':1,'j':[2,3]}] as a",
            [[{'i': 1, 'j': [2, 3]}, None, {'i': 1, 'j': [2, 3]}]],
        )

        # Struct that is NULL entirely
        df_duck = con.query("SELECT col0 as a FROM (VALUES ({'i':1,'j':2}), (NULL), ({'i':1,'j':2}), (NULL))").df()
        duck_values = df_duck['a']
        assert duck_values[0] == {'i': 1, 'j': 2}
        assert np.isnan(duck_values[1])
        assert duck_values[2] == {'i': 1, 'j': 2}
        assert np.isnan(duck_values[3])

        # MAP that is NULL entirely
        df_duck = con.query(
            "SELECT col0 as a FROM (VALUES (MAP(LIST_VALUE(1,2),LIST_VALUE(3,4))),(NULL), (MAP(LIST_VALUE(1,2),LIST_VALUE(3,4))), (NULL))"
        ).df()
        duck_values = df_duck['a']
        assert duck_values[0] == {'key': [1, 2], 'value': [3, 4]}
        assert np.isnan(duck_values[1])
        assert duck_values[2] == {'key': [1, 2], 'value': [3, 4]}
        assert np.isnan(duck_values[3])



import duckdb
import os
import pandas as pd
import pytest
from typing import Union, Optional
import sys

from duckdb.typing import (
    SQLNULL,
    BOOLEAN,
    TINYINT,
    UTINYINT,
    SMALLINT,
    USMALLINT,
    INTEGER,
    UINTEGER,
    BIGINT,
    UBIGINT,
    HUGEINT,
    UHUGEINT,
    UUID,
    FLOAT,
    DOUBLE,
    DATE,
    TIMESTAMP,
    TIMESTAMP_MS,
    TIMESTAMP_NS,
    TIMESTAMP_S,
    TIME,
    TIME_TZ,
    TIMESTAMP_TZ,
    VARCHAR,
    BLOB,
    BIT,
    INTERVAL,
)
import duckdb.typing


class TestType(object):
    def test_sqltype(self):
        assert str(duckdb.sqltype('struct(a VARCHAR, b BIGINT)')) == 'STRUCT(a VARCHAR, b BIGINT)'
        # todo: add tests with invalid type_str

    def test_primitive_types(self):
        assert str(SQLNULL) == 'NULL'
        assert str(BOOLEAN) == 'BOOLEAN'
        assert str(TINYINT) == 'TINYINT'
        assert str(UTINYINT) == 'UTINYINT'
        assert str(SMALLINT) == 'SMALLINT'
        assert str(USMALLINT) == 'USMALLINT'
        assert str(INTEGER) == 'INTEGER'
        assert str(UINTEGER) == 'UINTEGER'
        assert str(BIGINT) == 'BIGINT'
        assert str(UBIGINT) == 'UBIGINT'
        assert str(HUGEINT) == 'HUGEINT'
        assert str(UHUGEINT) == 'UHUGEINT'
        assert str(UUID) == 'UUID'
        assert str(FLOAT) == 'FLOAT'
        assert str(DOUBLE) == 'DOUBLE'
        assert str(DATE) == 'DATE'
        assert str(TIMESTAMP) == 'TIMESTAMP'
        assert str(TIMESTAMP_MS) == 'TIMESTAMP_MS'
        assert str(TIMESTAMP_NS) == 'TIMESTAMP_NS'
        assert str(TIMESTAMP_S) == 'TIMESTAMP_S'
        assert str(TIME) == 'TIME'
        assert str(TIME_TZ) == 'TIME WITH TIME ZONE'
        assert str(TIMESTAMP_TZ) == 'TIMESTAMP WITH TIME ZONE'
        assert str(VARCHAR) == 'VARCHAR'
        assert str(BLOB) == 'BLOB'
        assert str(BIT) == 'BIT'
        assert str(INTERVAL) == 'INTERVAL'

    def test_array_type(self):
        type = duckdb.array_type(BIGINT)
        assert str(type) == 'BIGINT[]'

    def test_struct_type(self):
        type = duckdb.struct_type({'a': BIGINT, 'b': BOOLEAN})
        assert str(type) == 'STRUCT(a BIGINT, b BOOLEAN)'

        # FIXME: create an unnamed struct when fields are provided as a list
        type = duckdb.struct_type([BIGINT, BOOLEAN])
        assert str(type) == 'STRUCT(v1 BIGINT, v2 BOOLEAN)'

    def test_map_type(self):
        type = duckdb.map_type(duckdb.sqltype("BIGINT"), duckdb.sqltype("DECIMAL(10, 2)"))
        assert str(type) == 'MAP(BIGINT, DECIMAL(10,2))'

    def test_decimal_type(self):
        type = duckdb.decimal_type(5, 3)
        assert str(type) == 'DECIMAL(5,3)'

    def test_string_type(self):
        type = duckdb.string_type()
        assert str(type) == 'VARCHAR'

    def test_string_type_collation(self):
        type = duckdb.string_type('NOCASE')
        # collation does not show up in the string representation..
        assert str(type) == 'VARCHAR'

    def test_union_type(self):
        type = duckdb.union_type([BIGINT, VARCHAR, TINYINT])
        assert str(type) == 'UNION(v1 BIGINT, v2 VARCHAR, v3 TINYINT)'

        type = duckdb.union_type({'a': BIGINT, 'b': VARCHAR, 'c': TINYINT})
        assert str(type) == 'UNION(a BIGINT, b VARCHAR, c TINYINT)'

    import sys

    @pytest.mark.skipif(sys.version_info < (3, 9), reason="requires >= python3.9")
    def test_implicit_convert_from_builtin_type(self):
        type = duckdb.list_type(list[str])
        assert str(type.child) == "VARCHAR[]"

        mapping = {'VARCHAR': str, 'BIGINT': int, 'BLOB': bytes, 'BLOB': bytearray, 'BOOLEAN': bool, 'DOUBLE': float}
        for expected, type in mapping.items():
            res = duckdb.list_type(type)
            assert str(res.child) == expected

        res = duckdb.list_type({'a': str, 'b': int})
        assert str(res.child) == 'STRUCT(a VARCHAR, b BIGINT)'

        res = duckdb.list_type(dict[str, int])
        assert str(res.child) == 'MAP(VARCHAR, BIGINT)'

        res = duckdb.list_type(list[str])
        assert str(res.child) == 'VARCHAR[]'

        res = duckdb.list_type(list[dict[str, dict[list[str], str]]])
        assert str(res.child) == 'MAP(VARCHAR, MAP(VARCHAR[], VARCHAR))[]'

        res = duckdb.list_type(list[Union[str, int]])
        assert str(res.child) == 'UNION(u1 VARCHAR, u2 BIGINT)[]'

    def test_implicit_convert_from_numpy(self, duckdb_cursor):
        np = pytest.importorskip("numpy")

        type_mapping = {
            'bool': 'BOOLEAN',
            'int8': 'TINYINT',
            'uint8': 'UTINYINT',
            'int16': 'SMALLINT',
            'uint16': 'USMALLINT',
            'int32': 'INTEGER',
            'uint32': 'UINTEGER',
            'int64': 'BIGINT',
            'uint64': 'UBIGINT',
            'float16': 'FLOAT',
            'float32': 'FLOAT',
            'float64': 'DOUBLE',
        }

        builtins = []
        builtins += [np.bool_]
        builtins += [np.byte]
        builtins += [np.ubyte]
        builtins += [np.short]
        builtins += [np.ushort]
        builtins += [np.intc]
        builtins += [np.uintc]
        builtins += [np.int_]
        builtins += [np.uint]
        builtins += [np.longlong]
        builtins += [np.ulonglong]
        builtins += [np.half]
        builtins += [np.float16]
        builtins += [np.single]
        builtins += [np.double]

        for builtin in builtins:
            print(builtin)
            type = duckdb_cursor.list_type(builtin)
            dtype_str = str(builtin().dtype)
            duckdb_type_str = str(type.child)
            assert type_mapping[dtype_str] == duckdb_type_str

    def test_attribute_accessor(self):
        type = duckdb.row_type([BIGINT, duckdb.list_type(duckdb.map_type(BLOB, BIT))])
        assert hasattr(type, 'a') == False
        assert hasattr(type, 'v1') == True

        field_one = type['v1']
        assert str(field_one) == 'BIGINT'
        field_one = type.v1
        assert str(field_one) == 'BIGINT'

        field_two = type['v2']
        assert str(field_two) == 'MAP(BLOB, BIT)[]'

        child_type = type.v2.child
        assert str(child_type) == 'MAP(BLOB, BIT)'

    # NOTE: we can support this, but I don't think going through hoops for an outdated version of python is worth it
    @pytest.mark.skipif(sys.version_info < (3, 9), reason="python3.7 does not store Optional[..] in a recognized way")
    def test_optional(self):
        type = duckdb.typing.DuckDBPyType(Optional[str])
        assert type == 'VARCHAR'
        type = duckdb.typing.DuckDBPyType(Optional[Union[int, bool]])
        assert type == 'UNION(u1 BIGINT, u2 BOOLEAN)'
        type = duckdb.typing.DuckDBPyType(Optional[list[int]])
        assert type == 'BIGINT[]'
        type = duckdb.typing.DuckDBPyType(Optional[dict[int, str]])
        assert type == 'MAP(BIGINT, VARCHAR)'
        type = duckdb.typing.DuckDBPyType(Optional[dict[Optional[int], Optional[str]]])
        assert type == 'MAP(BIGINT, VARCHAR)'
        type = duckdb.typing.DuckDBPyType(Optional[dict[Optional[int], Optional[str]]])
        assert type == 'MAP(BIGINT, VARCHAR)'
        type = duckdb.typing.DuckDBPyType(Optional[Union[Optional[str], Optional[bool]]])
        assert type == 'UNION(u1 VARCHAR, u2 BOOLEAN)'
        type = duckdb.typing.DuckDBPyType(Union[str, None])
        assert type == 'VARCHAR'

    @pytest.mark.skipif(sys.version_info < (3, 10), reason="'str | None' syntax requires Python 3.10 or higher")
    def test_optional_310(self):
        type = duckdb.typing.DuckDBPyType(str | None)
        assert type == 'VARCHAR'



import duckdb
import pandas as pd


class TestPandasUpdateList(object):
    def test_pandas_update_list(self, duckdb_cursor):
        duckdb_cursor = duckdb.connect(':memory:')
        duckdb_cursor.execute('create table t (l int[])')
        duckdb_cursor.execute('insert into t values ([1, 2]), ([3,4])')
        duckdb_cursor.execute('update t set l = [5, 6]')
        assert duckdb_cursor.execute('select * from t').fetchdf()['l'].tolist() == [[5, 6], [5, 6]]



import pytest
import duckdb


class TestRelationToView(object):
    def test_values_to_view(self, duckdb_cursor):
        rel = duckdb_cursor.values(['test', 'this is a long string'])
        res = rel.fetchall()
        assert res == [('test', 'this is a long string')]

        rel.to_view('vw1')

        view = duckdb_cursor.table('vw1')
        res = view.fetchall()
        assert res == [('test', 'this is a long string')]

    def test_relation_to_view(self, duckdb_cursor):
        rel = duckdb_cursor.sql("select 'test', 'this is a long string'")

        res = rel.fetchall()
        assert res == [('test', 'this is a long string')]

        rel.to_view('vw1')

        view = duckdb_cursor.table('vw1')
        res = view.fetchall()
        assert res == [('test', 'this is a long string')]

    def test_registered_relation(self, duckdb_cursor):
        rel = duckdb_cursor.sql("select 'test', 'this is a long string'")

        con = duckdb.connect()
        # Register on a different connection is not allowed
        with pytest.raises(
            duckdb.InvalidInputException,
            match='The relation you are attempting to register was not made from this connection',
        ):
            con.register('cross_connection', rel)

        # Register on the same connection just creates a view
        duckdb_cursor.register('same_connection', rel)
        view = duckdb_cursor.table('same_connection')
        res = view.fetchall()
        assert res == [('test', 'this is a long string')]



# cursor description

import duckdb
import tempfile
import os
import pytest


def check_exception(f):
    had_exception = False
    try:
        f()
    except BaseException:
        had_exception = True
    assert had_exception


class TestConnectionClose(object):
    def test_connection_close(self, duckdb_cursor):
        fd, db = tempfile.mkstemp()
        os.close(fd)
        os.remove(db)
        con = duckdb.connect(db)
        cursor = con.cursor()
        cursor.execute("create table a (i integer)")
        cursor.execute("insert into a values (42)")
        con.close()
        check_exception(lambda: cursor.execute("select * from a"))

    def test_open_and_exit(self):
        with pytest.raises(TypeError):
            with duckdb.connect() as connection:
                connection.execute("select 42")
                # This exception does not get swallowed by __exit__
                raise TypeError()

    def test_reopen_connection(self, duckdb_cursor):
        fd, db = tempfile.mkstemp()
        os.close(fd)
        os.remove(db)
        con = duckdb.connect(db)
        cursor = con.cursor()
        cursor.execute("create table a (i integer)")
        cursor.execute("insert into a values (42)")
        con.close()
        con = duckdb.connect(db)
        cursor = con.cursor()
        results = cursor.execute("select * from a").fetchall()
        assert results == [(42,)]



import duckdb


class TestModule:
    def test_paramstyle(self):
        assert duckdb.paramstyle == "qmark"

    def test_threadsafety(self):
        assert duckdb.threadsafety == 1

    def test_apilevel(self):
        assert duckdb.apilevel == "2.0"



import os

from mypy import stubtest

MYPY_INI_PATH = os.path.join(os.path.dirname(__file__), 'mypy.ini')


def test_generated_stubs():
    skip_stubs_errors = ['pybind11', 'git_revision', 'is inconsistent, metaclass differs']

    options = stubtest.parse_options(['duckdb', '--mypy-config-file', MYPY_INI_PATH])
    stubtest.test_stubs(options)

    broken_stubs = [
        error.get_description()
        for error in stubtest.test_module('duckdb')
        if not any(skip in error.get_description() for skip in skip_stubs_errors)
    ]

    if not broken_stubs:
        return
    print("Stubs must be updated, either add them to skip_stubs_errors or update __init__.pyi accordingly")
    print(broken_stubs)

    assert not broken_stubs



# simple DB API testcase

import duckdb
import pandas as pd
import pytest
from conftest import NumpyPandas, ArrowPandas
from packaging.version import Version

numpy_nullable_df = pd.DataFrame([{"COL1": "val1", "CoL2": 1.05}, {"COL1": "val4", "CoL2": 17}])

try:
    from pandas.compat import pa_version_under7p0

    pyarrow_dtypes_enabled = not pa_version_under7p0
except:
    pyarrow_dtypes_enabled = False

if Version(pd.__version__) >= Version('2.0.0') and pyarrow_dtypes_enabled:
    pyarrow_df = numpy_nullable_df.convert_dtypes(dtype_backend="pyarrow")
else:
    # dtype_backend is not supported in pandas < 2.0.0
    pyarrow_df = numpy_nullable_df


class TestImplicitPandasScan(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_local_pandas_scan(self, duckdb_cursor, pandas):
        con = duckdb.connect()
        df = pandas.DataFrame([{"COL1": "val1", "CoL2": 1.05}, {"COL1": "val3", "CoL2": 17}])
        r1 = con.execute('select * from df').fetchdf()
        assert r1["COL1"][0] == "val1"
        assert r1["COL1"][1] == "val3"
        assert r1["CoL2"][0] == 1.05
        assert r1["CoL2"][1] == 17

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_global_pandas_scan(self, duckdb_cursor, pandas):
        con = duckdb.connect()
        r1 = con.execute(f'select * from {pandas.backend}_df').fetchdf()
        assert r1["COL1"][0] == "val1"
        assert r1["COL1"][1] == "val4"
        assert r1["CoL2"][0] == 1.05
        assert r1["CoL2"][1] == 17



from multiprocessing.sharedctypes import Value
import numpy
import datetime
import pandas
import pytest
import duckdb
from io import StringIO, BytesIO
from duckdb.typing import BIGINT, VARCHAR, INTEGER


def TestFile(name):
    import os

    filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'data', name)
    return filename


@pytest.fixture
def create_temp_csv(tmp_path):
    # Create temporary CSV files
    file1_content = """1
2
3"""
    file2_content = """4
5
6"""
    file1_path = tmp_path / "file1.csv"
    file2_path = tmp_path / "file2.csv"

    file1_path.write_text(file1_content)
    file2_path.write_text(file2_content)

    return file1_path, file2_path


class TestReadCSV(object):
    def test_using_connection_wrapper(self):
        rel = duckdb.read_csv(TestFile('category.csv'))
        res = rel.fetchone()
        print(res)
        assert res == (1, 'Action', datetime.datetime(2006, 2, 15, 4, 46, 27))

    def test_using_connection_wrapper_with_keyword(self):
        rel = duckdb.read_csv(TestFile('category.csv'), dtype={'category_id': 'string'})
        res = rel.fetchone()
        print(res)
        assert res == ('1', 'Action', datetime.datetime(2006, 2, 15, 4, 46, 27))

    def test_no_options(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('category.csv'))
        res = rel.fetchone()
        print(res)
        assert res == (1, 'Action', datetime.datetime(2006, 2, 15, 4, 46, 27))

    def test_dtype(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('category.csv'), dtype={'category_id': 'string'})
        res = rel.fetchone()
        print(res)
        assert res == ('1', 'Action', datetime.datetime(2006, 2, 15, 4, 46, 27))

    def test_dtype_as_list(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('category.csv'), dtype=['string'])
        res = rel.fetchone()
        print(res)
        assert res == ('1', 'Action', datetime.datetime(2006, 2, 15, 4, 46, 27))

        rel = duckdb_cursor.read_csv(TestFile('category.csv'), dtype=['double'])
        res = rel.fetchone()
        print(res)
        assert res == (1.0, 'Action', datetime.datetime(2006, 2, 15, 4, 46, 27))

    def test_sep(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('category.csv'), sep=" ")
        res = rel.fetchone()
        print(res)
        assert res == ('1|Action|2006-02-15', datetime.time(4, 46, 27))

    def test_delimiter(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('category.csv'), delimiter=" ")
        res = rel.fetchone()
        print(res)
        assert res == ('1|Action|2006-02-15', datetime.time(4, 46, 27))

    def test_delimiter_and_sep(self, duckdb_cursor):
        with pytest.raises(duckdb.InvalidInputException, match="read_csv takes either 'delimiter' or 'sep', not both"):
            rel = duckdb_cursor.read_csv(TestFile('category.csv'), delimiter=" ", sep=" ")

    def test_header_true(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('category.csv'), header=True)
        res = rel.fetchone()
        print(res)
        assert res == (1, 'Action', datetime.datetime(2006, 2, 15, 4, 46, 27))

    @pytest.mark.skip(reason="Issue #6011 needs to be fixed first, header=False doesn't work correctly")
    def test_header_false(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('category.csv'), header=False)

    def test_na_values(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('category.csv'), na_values='Action')
        res = rel.fetchone()
        print(res)
        assert res == (1, None, datetime.datetime(2006, 2, 15, 4, 46, 27))

    def test_skiprows(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('category.csv'), skiprows=1)
        res = rel.fetchone()
        print(res)
        assert res == (1, 'Action', datetime.datetime(2006, 2, 15, 4, 46, 27))

    # We want to detect this at bind time
    def test_compression_wrong(self, duckdb_cursor):
        with pytest.raises(duckdb.Error, match="Input is not a GZIP stream"):
            rel = duckdb_cursor.read_csv(TestFile('category.csv'), compression='gzip')

    def test_quotechar(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('unquote_without_delimiter.csv'), quotechar="")
        res = rel.fetchone()
        print(res)
        assert res == ('"AAA"BB',)

    def test_escapechar(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('quote_escape.csv'), escapechar=";")
        res = rel.limit(1, 1).fetchone()
        print(res)
        assert res == ('345', 'TEST6', '"text""2""text"')

    def test_encoding_wrong(self, duckdb_cursor):
        with pytest.raises(
            duckdb.BinderException, match="Copy is only supported for UTF-8 encoded files, ENCODING 'UTF-8'"
        ):
            rel = duckdb_cursor.read_csv(TestFile('quote_escape.csv'), encoding=";")

    def test_encoding_correct(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('quote_escape.csv'), encoding="UTF-8")
        res = rel.limit(1, 1).fetchone()
        print(res)
        assert res == (345, 'TEST6', 'text"2"text')

    def test_date_format_as_datetime(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('datetime.csv'))
        res = rel.fetchone()
        print(res)
        assert res == (
            123,
            'TEST2',
            datetime.time(12, 12, 12),
            datetime.date(2000, 1, 1),
            datetime.datetime(2000, 1, 1, 12, 12),
        )

    def test_date_format_as_date(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('datetime.csv'), date_format='%Y-%m-%d')
        res = rel.fetchone()
        print(res)
        assert res == (
            123,
            'TEST2',
            datetime.time(12, 12, 12),
            datetime.date(2000, 1, 1),
            datetime.datetime(2000, 1, 1, 12, 12),
        )

    def test_timestamp_format(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('datetime.csv'), timestamp_format='%Y-%m-%d %H:%M:%S')
        res = rel.fetchone()
        assert res == (
            123,
            'TEST2',
            datetime.time(12, 12, 12),
            datetime.date(2000, 1, 1),
            datetime.datetime(2000, 1, 1, 12, 12),
        )

    def test_sample_size_correct(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('problematic.csv'), header=True, sample_size=-1)
        res = rel.fetchone()
        print(res)
        assert res == ('1', '1', '1')

    def test_all_varchar(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('category.csv'), all_varchar=True)
        res = rel.fetchone()
        print(res)
        assert res == ('1', 'Action', '2006-02-15 04:46:27')

    def test_null_padding(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('nullpadding.csv'), null_padding=False)
        res = rel.fetchall()
        assert res == [
            ('# this file has a bunch of gunk at the top',),
            ('one,two,three,four',),
            ('1,a,alice',),
            ('2,b,bob',),
        ]

        rel = duckdb_cursor.read_csv(TestFile('nullpadding.csv'), null_padding=True)
        res = rel.fetchall()
        assert res == [('one', 'two', 'three', 'four'), ('1', 'a', 'alice', None), ('2', 'b', 'bob', None)]

        rel = duckdb.read_csv(TestFile('nullpadding.csv'), null_padding=False)
        res = rel.fetchall()
        assert res == [
            ('# this file has a bunch of gunk at the top',),
            ('one,two,three,four',),
            ('1,a,alice',),
            ('2,b,bob',),
        ]

        rel = duckdb.read_csv(TestFile('nullpadding.csv'), null_padding=True)
        res = rel.fetchall()
        assert res == [('one', 'two', 'three', 'four'), ('1', 'a', 'alice', None), ('2', 'b', 'bob', None)]

        rel = duckdb_cursor.from_csv_auto(TestFile('nullpadding.csv'), null_padding=False)
        res = rel.fetchall()
        assert res == [
            ('# this file has a bunch of gunk at the top',),
            ('one,two,three,four',),
            ('1,a,alice',),
            ('2,b,bob',),
        ]

        rel = duckdb_cursor.from_csv_auto(TestFile('nullpadding.csv'), null_padding=True)
        res = rel.fetchall()
        assert res == [
            ('one', 'two', 'three', 'four'),
            ('1', 'a', 'alice', None),
            ('2', 'b', 'bob', None),
        ]

    def test_normalize_names(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('category.csv'), normalize_names=False)
        df = rel.df()
        column_names = list(df.columns.values)
        # The names are not normalized, so they are capitalized
        assert 'CATEGORY_ID' in column_names

        rel = duckdb_cursor.read_csv(TestFile('category.csv'), normalize_names=True)
        df = rel.df()
        column_names = list(df.columns.values)
        # The capitalized names are normalized to lowercase instead
        assert 'CATEGORY_ID' not in column_names

    def test_filename(self, duckdb_cursor):
        rel = duckdb_cursor.read_csv(TestFile('category.csv'), filename=False)
        df = rel.df()
        column_names = list(df.columns.values)
        # The filename is not included in the returned columns
        assert 'filename' not in column_names

        rel = duckdb_cursor.read_csv(TestFile('category.csv'), filename=True)
        df = rel.df()
        column_names = list(df.columns.values)
        # The filename is included in the returned columns
        assert 'filename' in column_names

    def test_read_pathlib_path(self, duckdb_cursor):
        pathlib = pytest.importorskip("pathlib")
        path = pathlib.Path(TestFile('category.csv'))
        rel = duckdb_cursor.read_csv(path)
        res = rel.fetchone()
        print(res)
        assert res == (1, 'Action', datetime.datetime(2006, 2, 15, 4, 46, 27))

    def test_read_filelike(self, duckdb_cursor):
        _ = pytest.importorskip("fsspec")
        string = StringIO("c1,c2,c3\na,b,c")
        res = duckdb_cursor.read_csv(string, header=True).fetchall()
        assert res == [('a', 'b', 'c')]

    def test_read_filelike_rel_out_of_scope(self, duckdb_cursor):
        _ = pytest.importorskip("fsspec")

        def keep_in_scope():
            string = StringIO("c1,c2,c3\na,b,c")
            # Create a ReadCSVRelation on a file-like object
            # this will add the object to our internal object filesystem
            rel = duckdb_cursor.read_csv(string, header=True)
            # The file-like object will still exist, so we can execute this later
            return rel

        def close_scope():
            string = StringIO("c1,c2,c3\na,b,c")
            # Create a ReadCSVRelation on a file-like object
            # this will add the object to our internal object filesystem
            res = duckdb_cursor.read_csv(string, header=True).fetchall()
            # When the relation goes out of scope - we delete the file-like object from our filesystem
            return res

        relation = keep_in_scope()
        res = relation.fetchall()

        res2 = close_scope()
        assert res == res2

    def test_filelike_bytesio(self, duckdb_cursor):
        _ = pytest.importorskip("fsspec")
        string = BytesIO(b"c1,c2,c3\na,b,c")
        res = duckdb_cursor.read_csv(string, header=True).fetchall()
        assert res == [('a', 'b', 'c')]

    def test_filelike_exception(self, duckdb_cursor):
        _ = pytest.importorskip("fsspec")

        class ReadError:
            def __init__(self):
                pass

            def read(self, amount):
                raise ValueError(amount)

            def seek(self, loc):
                return 0

        class SeekError:
            def __init__(self):
                pass

            def read(self, amount):
                return b'test'

            def seek(self, loc):
                raise ValueError(loc)

        obj = ReadError()
        with pytest.raises(ValueError):
            res = duckdb_cursor.read_csv(obj, header=True).fetchall()

        obj = SeekError()
        with pytest.raises(ValueError):
            res = duckdb_cursor.read_csv(obj, header=True).fetchall()

    def test_filelike_custom(self, duckdb_cursor):
        _ = pytest.importorskip("fsspec")

        class CustomIO:
            def __init__(self):
                self.loc = 0
                pass

            def seek(self, loc):
                self.loc = loc
                return loc

            def read(self, amount):
                out = b"c1,c2,c3\na,b,c"[self.loc : self.loc + amount : 1]
                self.loc += amount
                return out

        obj = CustomIO()
        res = duckdb_cursor.read_csv(obj, header=True).fetchall()
        assert res == [('a', 'b', 'c')]

    def test_filelike_non_readable(self, duckdb_cursor):
        _ = pytest.importorskip("fsspec")
        obj = 5
        with pytest.raises(ValueError, match="Can not read from a non file-like object"):
            res = duckdb_cursor.read_csv(obj, header=True).fetchall()

    def test_filelike_none(self, duckdb_cursor):
        _ = pytest.importorskip("fsspec")
        obj = None
        with pytest.raises(ValueError, match="Can not read from a non file-like object"):
            res = duckdb_cursor.read_csv(obj, header=True).fetchall()

    @pytest.mark.skip(reason="depends on garbage collector behaviour, and sporadically breaks in CI")
    def test_internal_object_filesystem_cleanup(self, duckdb_cursor):
        _ = pytest.importorskip("fsspec")

        class CountedObject(StringIO):
            instance_count = 0

            def __init__(self, str):
                CountedObject.instance_count += 1
                super().__init__(str)

            def __del__(self):
                CountedObject.instance_count -= 1

        def scoped_objects(duckdb_cursor):
            obj = CountedObject("a,b,c")
            rel1 = duckdb_cursor.read_csv(obj)
            assert rel1.fetchall() == [
                (
                    'a',
                    'b',
                    'c',
                )
            ]
            assert CountedObject.instance_count == 1

            obj = CountedObject("a,b,c")
            rel2 = duckdb_cursor.read_csv(obj)
            assert rel2.fetchall() == [
                (
                    'a',
                    'b',
                    'c',
                )
            ]
            assert CountedObject.instance_count == 2

            obj = CountedObject("a,b,c")
            rel3 = duckdb_cursor.read_csv(obj)
            assert rel3.fetchall() == [
                (
                    'a',
                    'b',
                    'c',
                )
            ]
            assert CountedObject.instance_count == 3

        assert CountedObject.instance_count == 0
        scoped_objects(duckdb_cursor)
        assert CountedObject.instance_count == 0

    def test_read_csv_glob(self, tmp_path, create_temp_csv):
        file1_path, file2_path = create_temp_csv

        # Use the temporary file paths to read CSV files
        con = duckdb.connect()
        rel = con.read_csv(f'{tmp_path}/file*.csv')
        res = con.sql("select * from rel order by all").fetchall()
        assert res == [(1,), (2,), (3,), (4,), (5,), (6,)]

    def test_read_csv_combined(self, duckdb_cursor):
        CSV_FILE = TestFile('stress_test.csv')
        COLUMNS = {
            'result': 'VARCHAR',
            'table': 'BIGINT',
            '_time': 'TIMESTAMPTZ',
            '_measurement': 'VARCHAR',
            'bench_test': 'VARCHAR',
            'flight_id': 'VARCHAR',
            'flight_status': 'VARCHAR',
            'log_level': 'VARCHAR',
            'sys_uuid': 'VARCHAR',
            'message': 'VARCHAR',
        }

        rel = duckdb.read_csv(
            CSV_FILE, header=True, skiprows=1, delimiter=",", quotechar='"', escapechar="\\", dtype=COLUMNS
        )
        res = rel.fetchall()

        rel2 = duckdb_cursor.sql(rel.sql_query())
        res2 = rel2.fetchall()

        # Assert that the results are the same
        assert res == res2

        # And assert that the columns and types of the relations are the same
        assert rel.columns == rel2.columns
        assert rel.types == rel2.types

    def test_read_csv_names(self):
        con = duckdb.connect()
        file = StringIO('one,two,three,four\n1,2,3,4\n1,2,3,4\n1,2,3,4')
        rel = con.read_csv(file, names=['a', 'b', 'c'])
        assert rel.columns == ['a', 'b', 'c', 'four']

        with pytest.raises(duckdb.InvalidInputException, match="read_csv only accepts 'names' as a list of strings"):
            rel = con.read_csv(file, names=True)

        # Excessive columns is fine, just doesn't have any effect past the number of provided columns
        rel = con.read_csv(file, names=['a', 'b', 'c', 'd', 'e'])
        assert rel.columns == ['a', 'b', 'c', 'd']

        # Duplicates are not okay
        with pytest.raises(duckdb.BinderException, match="has duplicate column name"):
            rel = con.read_csv(file, names=['a', 'b', 'a', 'b'])
            assert rel.columns == ['a', 'b', 'a', 'b']

    def test_read_csv_names_mixed_with_dtypes(self):
        con = duckdb.connect()
        file = StringIO('one,two,three,four\n1,2,3,4\n1,2,3,4\n1,2,3,4')
        rel = con.read_csv(
            file,
            names=['a', 'b', 'c'],
            dtype={
                'a': int,
                'b': bool,
                'c': str,
            },
        )
        assert rel.columns == ['a', 'b', 'c', 'four']
        assert rel.types == ['BIGINT', 'BOOLEAN', 'VARCHAR', 'BIGINT']

        # dtypes and names dont match
        # FIXME: seems the order columns are named in this error is non-deterministic
        # so for now I'm excluding the list of columns from the expected error
        expected_error = """do not exist in the CSV File"""
        with pytest.raises(duckdb.BinderException, match=expected_error):
            rel = con.read_csv(
                file,
                names=['a', 'b', 'c'],
                dtype={
                    'd': int,
                    'e': bool,
                    'f': str,
                },
            )

    def test_read_csv_multi_file(self):
        con = duckdb.connect()
        file1 = StringIO('one,two,three,four\n1,2,3,4\n1,2,3,4\n1,2,3,4')
        file2 = StringIO('one,two,three,four\n5,6,7,8\n5,6,7,8\n5,6,7,8')
        file3 = StringIO('one,two,three,four\n9,10,11,12\n9,10,11,12\n9,10,11,12')
        files = [file1, file2, file3]
        rel = con.read_csv(files)
        res = rel.fetchall()
        assert res == [
            (1, 2, 3, 4),
            (1, 2, 3, 4),
            (1, 2, 3, 4),
            (5, 6, 7, 8),
            (5, 6, 7, 8),
            (5, 6, 7, 8),
            (9, 10, 11, 12),
            (9, 10, 11, 12),
            (9, 10, 11, 12),
        ]

    def test_read_csv_empty_list(self):
        con = duckdb.connect()
        files = []
        with pytest.raises(
            duckdb.InvalidInputException, match='Please provide a non-empty list of paths or file-like objects'
        ):
            rel = con.read_csv(files)
            res = rel.fetchall()

    def test_read_csv_list_invalid_path(self):
        con = duckdb.connect()
        files = [
            StringIO('one,two,three,four\n1,2,3,4\n1,2,3,4\n1,2,3,4'),
            'not_valid_path',
            StringIO('one,two,three,four\n9,10,11,12\n9,10,11,12\n9,10,11,12'),
        ]
        with pytest.raises(duckdb.IOException, match='No files found that match the pattern "not_valid_path"'):
            rel = con.read_csv(files)
            res = rel.fetchall()



import os
import pytest
import shutil
from os.path import abspath, join, dirname, normpath
import glob
import duckdb
import warnings
from importlib import import_module

try:
    # need to ignore warnings that might be thrown deep inside pandas's import tree (from dateutil in this case)
    warnings.simplefilter(action='ignore', category=DeprecationWarning)
    pandas = import_module('pandas')
    warnings.resetwarnings()

    pyarrow_dtype = getattr(pandas, 'ArrowDtype', None)
except ImportError:
    pandas = None
    pyarrow_dtype = None

# Check if pandas has arrow dtypes enabled
try:
    from pandas.compat import pa_version_under7p0

    pyarrow_dtypes_enabled = not pa_version_under7p0
except ImportError:
    pyarrow_dtypes_enabled = False


def import_pandas():
    if pandas:
        return pandas
    else:
        pytest.skip("Couldn't import pandas")


# https://docs.pytest.org/en/latest/example/simple.html#control-skipping-of-tests-according-to-command-line-option
# https://stackoverflow.com/a/47700320
def pytest_addoption(parser):
    parser.addoption("--skiplist", action="append", nargs="+", type=str, help="skip listed tests")


def pytest_collection_modifyitems(config, items):
    tests_to_skip = config.getoption("--skiplist")
    if not tests_to_skip:
        # --skiplist not given in cli, therefore move on
        return

    # Combine all the lists into one
    skipped_tests = []
    for item in tests_to_skip:
        skipped_tests.extend(item)

    skip_listed = pytest.mark.skip(reason="included in --skiplist")
    for item in items:
        if item.name in skipped_tests:
            # test is named specifically
            item.add_marker(skip_listed)
        elif item.parent != None and item.parent.name in skipped_tests:
            # the class is named specifically
            item.add_marker(skip_listed)


@pytest.fixture(scope="function")
def duckdb_empty_cursor(request):
    connection = duckdb.connect('')
    cursor = connection.cursor()
    return cursor


def getTimeSeriesData(nper=None, freq: "Frequency" = "B"):
    from pandas import DatetimeIndex, bdate_range, Series
    from datetime import datetime
    from pandas._typing import Frequency
    import numpy as np
    import string

    _N = 30
    _K = 4

    def getCols(k) -> str:
        return string.ascii_uppercase[:k]

    def makeDateIndex(k: int = 10, freq: Frequency = "B", name=None, **kwargs) -> DatetimeIndex:
        dt = datetime(2000, 1, 1)
        dr = bdate_range(dt, periods=k, freq=freq, name=name)
        return DatetimeIndex(dr, name=name, **kwargs)

    def makeTimeSeries(nper=None, freq: Frequency = "B", name=None) -> Series:
        if nper is None:
            nper = _N
        return Series(np.random.randn(nper), index=makeDateIndex(nper, freq=freq), name=name)

    return {c: makeTimeSeries(nper, freq) for c in getCols(_K)}


def pandas_2_or_higher():
    from packaging.version import Version

    return Version(import_pandas().__version__) >= Version('2.0.0')


def pandas_supports_arrow_backend():
    try:
        from pandas.compat import pa_version_under11p0

        if pa_version_under11p0 == True:
            return False
    except ImportError:
        return False
    return pandas_2_or_higher()


def numpy_pandas_df(*args, **kwargs):
    return import_pandas().DataFrame(*args, **kwargs)


def arrow_pandas_df(*args, **kwargs):
    df = numpy_pandas_df(*args, **kwargs)
    return df.convert_dtypes(dtype_backend="pyarrow")


class NumpyPandas:
    def __init__(self):
        self.backend = 'numpy_nullable'
        self.DataFrame = numpy_pandas_df
        self.pandas = import_pandas()

    def __getattr__(self, name: str):
        return getattr(self.pandas, name)


def convert_arrow_to_numpy_backend(df):
    names = df.columns
    df_content = {}
    for name in names:
        df_content[name] = df[name].array.__arrow_array__()
    # This should convert the pyarrow chunked arrays into numpy arrays
    return import_pandas().DataFrame(df_content)


def convert_to_numpy(df):
    if (
        pyarrow_dtypes_enabled
        and pyarrow_dtype != None
        and any([True for x in df.dtypes if isinstance(x, pyarrow_dtype)])
    ):
        return convert_arrow_to_numpy_backend(df)
    return df


def convert_and_equal(df1, df2, **kwargs):
    df1 = convert_to_numpy(df1)
    df2 = convert_to_numpy(df2)
    import_pandas().testing.assert_frame_equal(df1, df2, **kwargs)


class ArrowMockTesting:
    def __init__(self):
        self.testing = import_pandas().testing
        self.assert_frame_equal = convert_and_equal

    def __getattr__(self, name: str):
        return getattr(self.testing, name)


# This converts dataframes constructed with 'DataFrame(...)' to pyarrow backed dataframes
# Assert equal does the opposite, turning all pyarrow backed dataframes into numpy backed ones
# this is done because we don't produce pyarrow backed dataframes yet
class ArrowPandas:
    def __init__(self):
        self.pandas = import_pandas()
        if pandas_2_or_higher() and pyarrow_dtypes_enabled:
            self.backend = 'pyarrow'
            self.DataFrame = arrow_pandas_df
        else:
            # For backwards compatible reasons, just mock regular pandas
            self.backend = 'numpy_nullable'
            self.DataFrame = self.pandas.DataFrame
        self.testing = ArrowMockTesting()

    def __getattr__(self, name: str):
        return getattr(self.pandas, name)


@pytest.fixture(scope="function")
def require():
    def _require(extension_name, db_name=''):
        # Paths to search for extensions

        build = normpath(join(dirname(__file__), "../../../build/"))
        extension = "extension/*/*.duckdb_extension"

        extension_search_patterns = [
            join(build, "release", extension),
            join(build, "debug", extension),
        ]

        # DUCKDB_PYTHON_TEST_EXTENSION_PATH can be used to add a path for the extension test to search for extensions
        if 'DUCKDB_PYTHON_TEST_EXTENSION_PATH' in os.environ:
            env_extension_path = os.getenv('DUCKDB_PYTHON_TEST_EXTENSION_PATH')
            env_extension_path = env_extension_path.rstrip('/')
            extension_search_patterns.append(env_extension_path + '/*/*.duckdb_extension')
            extension_search_patterns.append(env_extension_path + '/*.duckdb_extension')

        extension_paths_found = []
        for pattern in extension_search_patterns:
            extension_pattern_abs = abspath(pattern)
            print(f"Searching path: {extension_pattern_abs}")
            for path in glob.glob(extension_pattern_abs):
                extension_paths_found.append(path)

        for path in extension_paths_found:
            print(path)
            if path.endswith(extension_name + ".duckdb_extension"):
                conn = duckdb.connect(db_name, config={'allow_unsigned_extensions': 'true'})
                conn.execute(f"LOAD '{path}'")
                return conn
        pytest.skip(f'could not load {extension_name}')

    return _require


# By making the scope 'function' we ensure that a new connection gets created for every function that uses the fixture
@pytest.fixture(scope='function')
def spark():
    if not hasattr(spark, 'session'):
        # Cache the import
        from duckdb.experimental.spark.sql import SparkSession as session

        spark.session = session
    return spark.session.builder.master(':memory:').appName('pyspark').getOrCreate()


@pytest.fixture(scope='function')
def duckdb_cursor():
    connection = duckdb.connect('')
    yield connection
    connection.close()


@pytest.fixture(scope='function')
def integers(duckdb_cursor):
    cursor = duckdb_cursor
    cursor.execute('CREATE TABLE integers (i integer)')
    cursor.execute('INSERT INTO integers VALUES (0),(1),(2),(3),(4),(5),(6),(7),(8),(9),(NULL)')
    yield
    cursor.execute("drop table integers")


@pytest.fixture(scope='function')
def timestamps(duckdb_cursor):
    cursor = duckdb_cursor
    cursor.execute('CREATE TABLE timestamps (t timestamp)')
    cursor.execute("INSERT INTO timestamps VALUES ('1992-10-03 18:34:45'), ('2010-01-01 00:00:01'), (NULL)")
    yield
    cursor.execute("drop table timestamps")


@pytest.fixture(scope="function")
def duckdb_cursor_autocommit(request, tmp_path):
    test_dbfarm = tmp_path.resolve().as_posix()

    def finalizer():
        duckdb.shutdown()
        if tmp_path.is_dir():
            shutil.rmtree(test_dbfarm)

    request.addfinalizer(finalizer)

    connection = duckdb.connect(test_dbfarm)
    connection.set_autocommit(True)
    cursor = connection.cursor()
    return (cursor, connection, test_dbfarm)


@pytest.fixture(scope="function")
def initialize_duckdb(request, tmp_path):
    test_dbfarm = tmp_path.resolve().as_posix()

    def finalizer():
        duckdb.shutdown()
        if tmp_path.is_dir():
            shutil.rmtree(test_dbfarm)

    request.addfinalizer(finalizer)

    duckdb.connect(test_dbfarm)
    return test_dbfarm



import pandas as pd
import duckdb
import datetime
import numpy as np
import random


class TestPandasObject(object):
    def test_object_to_string(self, duckdb_cursor):
        con = duckdb.connect(database=':memory:', read_only=False)
        x = pd.DataFrame([[1, 'a', 2], [1, None, 2], [1, 1.1, 2], [1, 1.1, 2], [1, 1.1, 2]])
        x = x.iloc[1:].copy()  # middle col now entirely native float items
        con.register('view2', x)
        df = con.execute('select * from view2').fetchall()
        assert df == [(1, None, 2), (1, 1.1, 2), (1, 1.1, 2), (1, 1.1, 2)]

    def test_tuple_to_list(self, duckdb_cursor):
        tuple_df = pd.DataFrame.from_dict(
            dict(
                nums=[
                    (
                        1,
                        2,
                        3,
                    ),
                    (
                        4,
                        5,
                        6,
                    ),
                ]
            )
        )
        duckdb_cursor.execute("CREATE TABLE test as SELECT * FROM tuple_df")
        res = duckdb_cursor.table('test').fetchall()
        assert res == [([1, 2, 3],), ([4, 5, 6],)]

    def test_2273(self, duckdb_cursor):
        df_in = pd.DataFrame([[datetime.date(1992, 7, 30)]])
        assert duckdb_cursor.query("Select * from df_in").fetchall() == [(datetime.date(1992, 7, 30),)]

    def test_object_to_string_with_stride(self, duckdb_cursor):
        data = np.array([["a", "b", "c"], [1, 2, 3], [1, 2, 3], [11, 22, 33]])
        df = pd.DataFrame(data=data[1:,], columns=data[0])
        duckdb_cursor.register("object_with_strides", df)
        res = duckdb_cursor.sql('select * from object_with_strides').fetchall()
        assert res == [('1', '2', '3'), ('1', '2', '3'), ('11', '22', '33')]

    def test_2499(self, duckdb_cursor):
        df = pd.DataFrame(
            [
                [
                    np.array(
                        [
                            {'a': 0.881040697801939},
                            {'a': 0.9922600577751953},
                            {'a': 0.1589674833259317},
                            {'a': 0.8928451262745073},
                            {'a': 0.07022897889168278},
                        ],
                        dtype=object,
                    )
                ],
                [
                    np.array(
                        [
                            {'a': 0.8759413504156746},
                            {'a': 0.055784331256246156},
                            {'a': 0.8605151517439655},
                            {'a': 0.40807139339337695},
                            {'a': 0.8429048322459952},
                        ],
                        dtype=object,
                    )
                ],
                [
                    np.array(
                        [
                            {'a': 0.9697093934032401},
                            {'a': 0.9529257667149468},
                            {'a': 0.21398182248591713},
                            {'a': 0.6328512122275955},
                            {'a': 0.5146953214092728},
                        ],
                        dtype=object,
                    )
                ],
            ],
            columns=['col'],
        )

        con = duckdb.connect(database=':memory:', read_only=False)
        con.register('df', df)
        assert con.execute('select count(*) from df').fetchone() == (3,)



import duckdb
import pytest


@pytest.fixture
def con():
    conn = duckdb.connect()
    conn.execute(
        """
		create table tbl as (SELECT * FROM (VALUES
			(1, 'a', 12),
			(1, 'a', 10),
			(2, 'b', 5),
			(2, 'a', 7),
			(3, 'a', 5),
			(5, 'c', 2)
		) AS tbl(a, b, c))
	"""
    )
    yield conn


class TestGroupings(object):
    def test_basic_grouping(self, con):
        rel = con.table('tbl').sum("a", "b")
        res = rel.fetchall()
        assert res == [(7,), (2,), (5,)]

        rel = con.sql("select sum(a) from tbl GROUP BY b")
        res2 = rel.fetchall()
        assert res == res2

    def test_cubed(self, con):
        rel = con.table('tbl').sum("a", "CUBE (b)").order("ALL")
        res = rel.fetchall()
        assert res == [(2,), (5,), (7,), (14,)]

        rel = con.sql("select sum(a) from tbl GROUP BY CUBE (b) ORDER BY ALL")
        res2 = rel.fetchall()
        assert res == res2

    def test_rollup(self, con):
        rel = con.table('tbl').sum("a", "ROLLUP (b, c)").order("ALL")
        res = rel.fetchall()
        assert res == [(1,), (1,), (2,), (2,), (2,), (3,), (5,), (5,), (7,), (14,)]

        rel = con.sql("select sum(a) from tbl GROUP BY ROLLUP (b, c) ORDER BY ALL")
        res2 = rel.fetchall()
        assert res == res2



import duckdb
import os
import pytest

pd = pytest.importorskip("pandas")
pa = pytest.importorskip("pyarrow")
from typing import Union
import pyarrow.compute as pc
import uuid
import datetime

from duckdb.typing import *


class TestPyArrowUDF(object):
    def test_basic_use(self):
        def plus_one(x):
            table = pa.lib.Table.from_arrays([x], names=['c0'])
            import pandas as pd

            df = pd.DataFrame(x.to_pandas())
            df['c0'] = df['c0'] + 1
            return pa.lib.Table.from_pandas(df)

        con = duckdb.connect()
        con.create_function('plus_one', plus_one, [BIGINT], BIGINT, type='arrow')
        assert [(6,)] == con.sql('select plus_one(5)').fetchall()

        range_table = con.table_function('range', [5000])
        res = con.sql('select plus_one(i) from range_table tbl(i)').fetchall()
        assert len(res) == 5000

        vector_size = duckdb.__standard_vector_size__
        res = con.sql(f'select i, plus_one(i) from test_vector_types(NULL::BIGINT, false) t(i), range({vector_size})')
        assert len(res) == (vector_size * 11)

    # NOTE: This only works up to duckdb.__standard_vector_size__,
    # because we process up to STANDARD_VECTOR_SIZE tuples at a time
    def test_sort_table(self):
        def sort_table(x):
            table = pa.lib.Table.from_arrays([x], names=['c0'])
            sorted_table = table.sort_by([("c0", "ascending")])
            return sorted_table

        con = duckdb.connect()
        con.create_function('sort_table', sort_table, [BIGINT], BIGINT, type='arrow')
        res = con.sql("select 100-i as original, sort_table(original) from range(100) tbl(i)").fetchall()
        assert res[0] == (100, 1)

    def test_varargs(self):
        def variable_args(*args):
            # We return a chunked array here, but internally we convert this into a Table
            if len(args) == 0:
                raise ValueError("Expected at least one argument")
            for item in args:
                return item

        con = duckdb.connect()
        # This function takes any number of arguments, returning the first column
        con.create_function('varargs', variable_args, None, BIGINT, type='arrow')
        res = con.sql("""select varargs(5, '3', '2', 1, 0.12345)""").fetchall()
        assert res == [(5,)]

        res = con.sql("""select varargs(42, 'test', [5,4,3])""").fetchall()
        assert res == [(42,)]

    def test_cast_varchar_to_int(self):
        def takes_string(col):
            return col

        con = duckdb.connect()
        # The return type of the function is set to BIGINT, but it takes a VARCHAR
        con.create_function('pyarrow_string_to_num', takes_string, [VARCHAR], BIGINT, type='arrow')

        # Succesful conversion
        res = con.sql("""select pyarrow_string_to_num('5')""").fetchall()
        assert res == [(5,)]

        with pytest.raises(duckdb.ConversionException, match="""Could not convert string 'test' to INT64"""):
            res = con.sql("""select pyarrow_string_to_num('test')""").fetchall()

    def test_return_multiple_columns(self):
        def returns_two_columns(col):
            import pandas as pd

            # Return a pyarrow table consisting of two columns
            return pa.lib.Table.from_pandas(pd.DataFrame({'a': [5, 4, 3], 'b': ['test', 'quack', 'duckdb']}))

        con = duckdb.connect()
        # Scalar functions only return a single value per tuple
        con.create_function('two_columns', returns_two_columns, [BIGINT], BIGINT, type='arrow')
        with pytest.raises(
            duckdb.InvalidInputException,
            match='The returned table from a pyarrow scalar udf should only contain one column, found 2',
        ):
            res = con.sql("""select two_columns(5)""").fetchall()

    def test_return_none(self):
        def returns_none(col):
            return None

        con = duckdb.connect()
        con.create_function('will_crash', returns_none, [BIGINT], BIGINT, type='arrow')
        with pytest.raises(duckdb.Error, match="""Could not convert the result into an Arrow Table"""):
            res = con.sql("""select will_crash(5)""").fetchall()

    def test_empty_result(self):
        def return_empty(col):
            # Always returns an empty table
            return pa.lib.Table.from_arrays([[]], names=['c0'])

        con = duckdb.connect()
        con.create_function('empty_result', return_empty, [BIGINT], BIGINT, type='arrow')
        with pytest.raises(duckdb.InvalidInputException, match='Returned pyarrow table should have 1 tuples, found 0'):
            res = con.sql("""select empty_result(5)""").fetchall()

    def test_excessive_result(self):
        def return_too_many(col):
            # Always returns a table consisting of 5 tuples
            return pa.lib.Table.from_arrays([[5, 4, 3, 2, 1]], names=['c0'])

        con = duckdb.connect()
        con.create_function('too_many_tuples', return_too_many, [BIGINT], BIGINT, type='arrow')
        with pytest.raises(duckdb.InvalidInputException, match='Returned pyarrow table should have 1 tuples, found 5'):
            res = con.sql("""select too_many_tuples(5)""").fetchall()

    def test_arrow_side_effects(self, duckdb_cursor):
        import random as r

        def random_arrow(x):
            if not hasattr(random_arrow, 'data'):
                random_arrow.data = 0

            input = x.to_pylist()
            val = random_arrow.data
            output = [val + i for i in range(len(input))]
            random_arrow.data += len(input)
            return output

        duckdb_cursor.create_function(
            "random_arrow",
            random_arrow,
            [VARCHAR],
            INTEGER,
            side_effects=True,
            type="arrow",
        )
        res = duckdb_cursor.query("SELECT random_arrow('') FROM range(10)").fetchall()
        assert res == [(0,), (1,), (2,), (3,), (4,), (5,), (6,), (7,), (8,), (9,)]

    def test_return_struct(self):
        def return_struct(col):
            con = duckdb.connect()
            return con.sql(
                """
                select {'a': 5, 'b': 'test', 'c': [5,3,2]}
            """
            ).arrow()

        con = duckdb.connect()
        struct_type = con.struct_type({'a': BIGINT, 'b': VARCHAR, 'c': con.list_type(BIGINT)})
        con.create_function('return_struct', return_struct, [BIGINT], struct_type, type='arrow')
        res = con.sql("""select return_struct(5)""").fetchall()
        assert res == [({'a': 5, 'b': 'test', 'c': [5, 3, 2]},)]

    def test_multiple_chunks(self):
        def return_unmodified(col):
            return col

        con = duckdb.connect()
        con.create_function('unmodified', return_unmodified, [BIGINT], BIGINT, type='arrow')
        res = con.sql(
            """
            select unmodified(i) from range(5000) tbl(i)
        """
        ).fetchall()

        assert len(res) == 5000
        assert res == con.sql('select * from range(5000)').fetchall()

    def test_inferred(self):
        def func(x: int) -> int:
            import pandas as pd

            df = pd.DataFrame({'c0': x})
            df['c0'] = df['c0'] ** 2
            return pa.lib.Table.from_pandas(df)

        con = duckdb.connect()
        con.create_function('inferred', func, type='arrow')
        res = con.sql('select inferred(42)').fetchall()
        assert res == [(1764,)]

    def test_nulls(self):
        def return_five(x):
            import pandas as pd

            length = len(x)
            return pa.lib.Table.from_pandas(pd.DataFrame({'a': [5 for _ in range(length)]}))

        con = duckdb.connect()
        con.create_function('return_five', return_five, [BIGINT], BIGINT, null_handling='special', type='arrow')
        res = con.sql('select return_five(NULL) from range(10)').fetchall()
        # without 'special' null handling these would all be NULL
        assert res == [(5,), (5,), (5,), (5,), (5,), (5,), (5,), (5,), (5,), (5,)]

        con = duckdb.connect()
        con.create_function('return_five', return_five, [BIGINT], BIGINT, null_handling='default', type='arrow')
        res = con.sql('select return_five(NULL) from range(10)').fetchall()
        # Because we didn't specify 'special' null handling, these are all NULL
        assert res == [(None,), (None,), (None,), (None,), (None,), (None,), (None,), (None,), (None,), (None,)]



class TestUnsigned(object):
    def test_unsigned(self, duckdb_cursor):
        duckdb_cursor.execute('create table unsigned (a utinyint, b usmallint, c uinteger, d ubigint)')
        duckdb_cursor.execute('insert into unsigned values (1,1,1,1), (null,null,null,null)')
        duckdb_cursor.execute('select * from unsigned order by a nulls first')
        result = duckdb_cursor.fetchall()
        assert result == [(None, None, None, None), (1, 1, 1, 1)]



# -*- coding: iso-8859-1 -*-
# pysqlite2/test/types.py: tests for type conversion and detection
#
# Copyright (C) 2005 Gerhard Häring <gh@ghaering.de>
#
# This file is part of pyduckdb.
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#
# This library is derived from the pysqlite testing library, with small modifications
# to remove tests that are for features that are not supported by DuckDB.

import datetime
import decimal
import unittest
import duckdb
import pytest


class DuckDBTypeTests(unittest.TestCase):
    def setUp(self):
        self.con = duckdb.connect(":memory:")
        self.cur = self.con.cursor()
        self.cur.execute("create table test(i bigint, s varchar, f double, b BLOB)")

    def tearDown(self):
        self.cur.close()
        self.con.close()

    def test_CheckString(self):
        self.cur.execute("insert into test(s) values (?)", (u"Österreich",))
        self.cur.execute("select s from test")
        row = self.cur.fetchone()
        self.assertEqual(row[0], u"Österreich")

    def test_CheckSmallInt(self):
        self.cur.execute("insert into test(i) values (?)", (42,))
        self.cur.execute("select i from test")
        row = self.cur.fetchone()
        self.assertEqual(row[0], 42)

    def test_CheckLargeInt(self):
        num = 2**40
        self.cur.execute("insert into test(i) values (?)", (num,))
        self.cur.execute("select i from test")
        row = self.cur.fetchone()
        self.assertEqual(row[0], num)

    def test_CheckFloat(self):
        val = 3.14
        self.cur.execute("insert into test(f) values (?)", (val,))
        self.cur.execute("select f from test")
        row = self.cur.fetchone()
        self.assertEqual(row[0], val)

    def test_CheckDecimalTooBig(self):
        val = 17.29
        self.cur.execute("insert into test(f) values (?)", (decimal.Decimal(val),))
        self.cur.execute("select f from test")
        row = self.cur.fetchone()
        self.assertEqual(row[0], val)

    def test_CheckDecimal(self):
        val = '17.29'
        val = decimal.Decimal(val)
        self.cur.execute("insert into test(f) values (?)", (val,))
        self.cur.execute("select f from test")
        row = self.cur.fetchone()
        self.assertEqual(row[0], self.cur.execute("select 17.29::DOUBLE").fetchone()[0])

    def test_CheckDecimalWithExponent(self):
        val = '1E5'
        val = decimal.Decimal(val)
        self.cur.execute("insert into test(f) values (?)", (val,))
        self.cur.execute("select f from test")
        row = self.cur.fetchone()
        self.assertEqual(row[0], self.cur.execute("select 1.00000::DOUBLE").fetchone()[0])

    def test_CheckNaN(self):
        import math

        val = decimal.Decimal('nan')
        self.cur.execute("insert into test(f) values (?)", (val,))
        self.cur.execute("select f from test")
        row = self.cur.fetchone()
        self.assertEqual(math.isnan(row[0]), True)

    def test_CheckInf(self):
        val = decimal.Decimal('inf')
        self.cur.execute("insert into test(f) values (?)", (val,))
        self.cur.execute("select f from test")
        row = self.cur.fetchone()
        self.assertEqual(row[0], val)

    def test_CheckBytesBlob(self):
        val = b"Guglhupf"
        self.cur.execute("insert into test(b) values (?)", (val,))
        self.cur.execute("select b from test")
        row = self.cur.fetchone()
        self.assertEqual(row[0], val)

    def test_CheckMemoryviewBlob(self):
        sample = b"Guglhupf"
        val = memoryview(sample)
        self.cur.execute("insert into test(b) values (?)", (val,))
        self.cur.execute("select b from test")
        row = self.cur.fetchone()
        self.assertEqual(row[0], sample)

    def test_CheckMemoryviewFromhexBlob(self):
        sample = bytes.fromhex('00FF0F2E3D4C5B6A798800FF00')
        val = memoryview(sample)
        self.cur.execute("insert into test(b) values (?)", (val,))
        self.cur.execute("select b from test")
        row = self.cur.fetchone()
        self.assertEqual(row[0], sample)

    def test_CheckNoneBlob(self):
        val = None
        self.cur.execute("insert into test(b) values (?)", (val,))
        self.cur.execute("select b from test")
        row = self.cur.fetchone()
        self.assertEqual(row[0], val)

    def test_CheckUnicodeExecute(self):
        self.cur.execute(u"select 'Österreich'")
        row = self.cur.fetchone()
        self.assertEqual(row[0], u"Österreich")


class CommonTableExpressionTests(unittest.TestCase):
    def setUp(self):
        self.con = duckdb.connect(":memory:")
        self.cur = self.con.cursor()
        self.cur.execute("create table test(x int)")

    def tearDown(self):
        self.cur.close()
        self.con.close()

    def test_CheckCursorDescriptionCTESimple(self):
        self.cur.execute("with one as (select 1) select * from one")
        self.assertIsNotNone(self.cur.description)
        self.assertEqual(self.cur.description[0][0], "1")

    def test_CheckCursorDescriptionCTESMultipleColumns(self):
        self.cur.execute("insert into test values(1)")
        self.cur.execute("insert into test values(2)")
        self.cur.execute("with testCTE as (select * from test) select * from testCTE")
        self.assertIsNotNone(self.cur.description)
        self.assertEqual(self.cur.description[0][0], "x")

    def test_CheckCursorDescriptionCTE(self):
        self.cur.execute("insert into test values (1)")
        self.cur.execute("with bar as (select * from test) select * from test where x = 1")
        self.assertIsNotNone(self.cur.description)
        self.assertEqual(self.cur.description[0][0], "x")
        self.cur.execute("with bar as (select * from test) select * from test where x = 2")
        self.assertIsNotNone(self.cur.description)
        self.assertEqual(self.cur.description[0][0], "x")


class DateTimeTests(unittest.TestCase):
    def setUp(self):
        self.con = duckdb.connect(":memory:")
        self.cur = self.con.cursor()
        self.cur.execute("create table test(d date, t time, ts timestamp)")

    def tearDown(self):
        self.cur.close()
        self.con.close()

    def test_CheckDate(self):
        d = datetime.date(2004, 2, 14)
        self.cur.execute("insert into test(d) values (?)", (d,))
        self.cur.execute("select d from test")
        d2 = self.cur.fetchone()[0]
        self.assertEqual(d, d2)

    def test_CheckTime(self):
        t = datetime.time(7, 15, 0)
        self.cur.execute("insert into test(t) values (?)", (t,))
        self.cur.execute("select t from test")
        t2 = self.cur.fetchone()[0]
        self.assertEqual(t, t2)

    def test_CheckTimestamp(self):
        ts = datetime.datetime(2004, 2, 14, 7, 15, 0)
        self.cur.execute("insert into test(ts) values (?)", (ts,))
        self.cur.execute("select ts from test")
        ts2 = self.cur.fetchone()[0]
        self.assertEqual(ts, ts2)

    def test_CheckSqlTimestamp(self):
        now = datetime.datetime.now(datetime.UTC) if hasattr(datetime, 'UTC') else datetime.datetime.utcnow()
        self.cur.execute("insert into test(ts) values (current_timestamp)")
        self.cur.execute("select ts from test")
        ts = self.cur.fetchone()[0]
        self.assertEqual(type(ts), datetime.datetime)
        self.assertEqual(ts.year, now.year)

    def test_CheckDateTimeSubSeconds(self):
        ts = datetime.datetime(2004, 2, 14, 7, 15, 0, 500000)
        self.cur.execute("insert into test(ts) values (?)", (ts,))
        self.cur.execute("select ts from test")
        ts2 = self.cur.fetchone()[0]
        self.assertEqual(ts, ts2)

    def test_CheckTimeSubSeconds(self):
        t = datetime.time(7, 15, 0, 500000)
        self.cur.execute("insert into test(t) values (?)", (t,))
        self.cur.execute("select t from test")
        t2 = self.cur.fetchone()[0]
        self.assertEqual(t, t2)

    def test_CheckDateTimeSubSecondsFloatingPoint(self):
        ts = datetime.datetime(2004, 2, 14, 7, 15, 0, 510241)
        self.cur.execute("insert into test(ts) values (?)", (ts,))
        self.cur.execute("select ts from test")
        ts2 = self.cur.fetchone()[0]
        self.assertEqual(ts.year, ts2.year)
        self.assertEqual(ts2.microsecond, 510241)


class ListTests(unittest.TestCase):
    def setUp(self):
        self.con = duckdb.connect(":memory:")
        self.cur = self.con.cursor()
        self.cur.execute("create table test(single INTEGER[], nested INTEGER[][])")

    def tearDown(self):
        self.cur.close()
        self.con.close()

    def test_CheckEmptyList(self):
        val = []
        self.cur.execute("insert into test values (?, ?)", (val, val))
        self.assertEqual(
            self.cur.execute("select * from test").fetchall(),
            [(val, val)],
        )

    def test_CheckSingleList(self):
        val = [1, 2, 3]
        self.cur.execute("insert into test(single) values (?)", (val,))
        self.assertEqual(
            self.cur.execute("select * from test").fetchall(),
            [(val, None)],
        )

    def test_CheckNestedList(self):
        val = [[1], [2], [3, 4]]
        self.cur.execute("insert into test(nested) values (?)", (val,))
        self.assertEqual(
            self.cur.execute("select * from test").fetchall(),
            [
                (
                    None,
                    val,
                )
            ],
        )

    def test_CheckNone(self):
        val = None
        self.cur.execute("insert into test values (?, ?)", (val, val))
        self.assertEqual(
            self.cur.execute("select * from test").fetchall(),
            [(val, val)],
        )

    def test_CheckEmbeddedNone(self):
        val = [None]
        self.cur.execute("insert into test values (?, ?)", (val, val))
        self.assertEqual(
            self.cur.execute("select * from test").fetchall(),
            [(val, val)],
        )



import duckdb
import datetime
import pytest


def create_query(positive, type):
    inf = 'infinity' if positive else '-infinity'
    return f"""
        select '{inf}'::{type}
    """


class TestDateTimeDateTime(object):
    @pytest.mark.parametrize('positive', [True, False])
    @pytest.mark.parametrize(
        'type',
        [
            'TIMESTAMP',
            'TIMESTAMP_S',
            'TIMESTAMP_MS',
            'TIMESTAMP_NS',
            'TIMESTAMPTZ',
            'TIMESTAMP_US',
        ],
    )
    def test_timestamp_infinity(self, positive, type):
        con = duckdb.connect()

        if type in ['TIMESTAMP_S', 'TIMESTAMP_MS', 'TIMESTAMP_NS']:
            # Infinity (both positive and negative) is not supported for non-usecond timetamps
            return

        expected_val = datetime.datetime.max if positive else datetime.datetime.min
        query = create_query(positive, type)
        res = con.sql(query).fetchall()[0][0]
        assert res == expected_val

    def test_timestamp_infinity_roundtrip(self):
        con = duckdb.connect()

        # positive infinity
        con.execute("select $1, $1 = 'infinity'::TIMESTAMP", [datetime.datetime.max])
        res = con.fetchall()
        assert res == [(datetime.datetime.max, False)]

        # negative infinity
        con.execute("select $1, $1 = '-infinity'::TIMESTAMP", [datetime.datetime.min])
        res = con.fetchall()
        assert res == [(datetime.datetime.min, False)]

    def test_convert_negative_interval(self, duckdb_cursor):
        res = duckdb_cursor.execute(
            "SELECT CAST('2023-07-22T11:28:07' AS TIMESTAMP) - CAST('2023-07-23T11:28:07' AS TIMESTAMP)"
        ).fetchall()
        assert res == [(datetime.timedelta(days=-1),)]



import duckdb
import pytest
import tempfile
import numpy
import pandas
import datetime

pa = pytest.importorskip("pyarrow")


def parquet_types_test(type_list):
    temp = tempfile.NamedTemporaryFile()
    temp_name = temp.name
    for type_pair in type_list:
        value_list = type_pair[0]
        numpy_type = type_pair[1]
        sql_type = type_pair[2]
        add_cast = len(type_pair) > 3 and type_pair[3]
        add_sql_cast = len(type_pair) > 4 and type_pair[4]
        df = pandas.DataFrame.from_dict({'val': numpy.array(value_list, dtype=numpy_type)})
        duckdb_cursor = duckdb.connect()
        duckdb_cursor.execute(f"CREATE TABLE tmp AS SELECT val::{sql_type} val FROM df")
        duckdb_cursor.execute(f"COPY tmp TO '{temp_name}' (FORMAT PARQUET)")
        read_df = pandas.read_parquet(temp_name)
        if add_cast:
            read_df['val'] = read_df['val'].astype(numpy_type)
        assert df.equals(read_df)

        read_from_duckdb = duckdb_cursor.execute(f"SELECT * FROM parquet_scan('{temp_name}')").df()
        assert read_df.equals(read_from_duckdb)

        df.to_parquet(temp_name)
        if add_sql_cast:
            read_from_arrow = duckdb_cursor.execute(f"SELECT val::{sql_type} val FROM parquet_scan('{temp_name}')").df()
        else:
            read_from_arrow = duckdb_cursor.execute(f"SELECT * FROM parquet_scan('{temp_name}')").df()
        assert read_df.equals(read_from_arrow)


class TestParquetRoundtrip(object):
    def test_roundtrip_numeric(self, duckdb_cursor):
        type_list = [
            ([-(2**7), 0, 2**7 - 1], numpy.int8, 'TINYINT'),
            ([-(2**15), 0, 2**15 - 1], numpy.int16, 'SMALLINT'),
            ([-(2**31), 0, 2**31 - 1], numpy.int32, 'INTEGER'),
            ([-(2**63), 0, 2**63 - 1], numpy.int64, 'BIGINT'),
            ([0, 42, 2**8 - 1], numpy.uint8, 'UTINYINT'),
            ([0, 42, 2**16 - 1], numpy.uint16, 'USMALLINT'),
            ([0, 42, 2**32 - 1], numpy.uint32, 'UINTEGER', False, True),
            ([0, 42, 2**64 - 1], numpy.uint64, 'UBIGINT'),
            ([0, 0.5, -0.5], numpy.float32, 'REAL'),
            ([0, 0.5, -0.5], numpy.float64, 'DOUBLE'),
        ]
        parquet_types_test(type_list)

    def test_roundtrip_timestamp(self, duckdb_cursor):
        date_time_list = [
            datetime.datetime(2018, 3, 10, 11, 17, 54),
            datetime.datetime(1900, 12, 12, 23, 48, 42),
            None,
            datetime.datetime(1992, 7, 9, 7, 5, 33),
        ]
        type_list = [
            (date_time_list, 'datetime64[ns]', 'TIMESTAMP_NS'),
            (date_time_list, 'datetime64[us]', 'TIMESTAMP'),
            (date_time_list, 'datetime64[ms]', 'TIMESTAMP_MS'),
            (date_time_list, 'datetime64[s]', 'TIMESTAMP_S'),
            (date_time_list, 'datetime64[D]', 'DATE', True),
        ]
        parquet_types_test(type_list)

    def test_roundtrip_varchar(self, duckdb_cursor):
        varchar_list = ['hello', 'this is a very long string', 'hello', None]
        type_list = [(varchar_list, object, 'VARCHAR')]
        parquet_types_test(type_list)



import duckdb
import pytest

pl = pytest.importorskip("polars")
arrow = pytest.importorskip("pyarrow")
pl_testing = pytest.importorskip("polars.testing")


class TestPolars(object):
    def test_polars(self, duckdb_cursor):
        df = pl.DataFrame(
            {
                "A": [1, 2, 3, 4, 5],
                "fruits": ["banana", "banana", "apple", "apple", "banana"],
                "B": [5, 4, 3, 2, 1],
                "cars": ["beetle", "audi", "beetle", "beetle", "beetle"],
            }
        )
        # scan plus return a polars dataframe
        polars_result = duckdb_cursor.sql('SELECT * FROM df').pl()
        pl_testing.assert_frame_equal(df, polars_result)

        # now do the same for a lazy dataframe
        lazy_df = df.lazy()
        lazy_result = duckdb_cursor.sql('SELECT * FROM lazy_df').pl()
        pl_testing.assert_frame_equal(df, lazy_result)

        con = duckdb.connect()
        con_result = con.execute('SELECT * FROM df').pl()
        pl_testing.assert_frame_equal(df, con_result)

    def test_register_polars(self, duckdb_cursor):
        con = duckdb.connect()
        df = pl.DataFrame(
            {
                "A": [1, 2, 3, 4, 5],
                "fruits": ["banana", "banana", "apple", "apple", "banana"],
                "B": [5, 4, 3, 2, 1],
                "cars": ["beetle", "audi", "beetle", "beetle", "beetle"],
            }
        )
        # scan plus return a polars dataframe
        con.register('polars_df', df)
        polars_result = con.execute('select * from polars_df').pl()
        pl_testing.assert_frame_equal(df, polars_result)
        con.unregister('polars_df')
        with pytest.raises(duckdb.CatalogException, match='Table with name polars_df does not exist'):
            con.execute("SELECT * FROM polars_df;").pl()

        con.register('polars_df', df.lazy())
        polars_result = con.execute('select * from polars_df').pl()
        pl_testing.assert_frame_equal(df, polars_result)



# simple DB API testcase


class TestSimpleDBAPI(object):
    def test_prepare(self, duckdb_cursor):
        result = duckdb_cursor.execute('SELECT CAST(? AS INTEGER), CAST(? AS INTEGER)', ['42', '84']).fetchall()
        assert result == [
            (
                42,
                84,
            )
        ], "Incorrect result returned"

        c = duckdb_cursor

        # from python docs
        c.execute(
            '''CREATE TABLE stocks
             (date text, trans text, symbol text, qty real, price real)'''
        )
        c.execute("INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14)")

        t = ('RHAT',)
        result = c.execute('SELECT COUNT(*) FROM stocks WHERE symbol=?', t).fetchone()
        assert result == (1,)

        t = ['RHAT']
        result = c.execute('SELECT COUNT(*) FROM stocks WHERE symbol=?', t).fetchone()
        assert result == (1,)

        # Larger example that inserts many records at a time
        purchases = [
            ('2006-03-28', 'BUY', 'IBM', 1000, 45.00),
            ('2006-04-05', 'BUY', 'MSFT', 1000, 72.00),
            ('2006-04-06', 'SELL', 'IBM', 500, 53.00),
        ]
        c.executemany('INSERT INTO stocks VALUES (?,?,?,?,?)', purchases)

        result = c.execute('SELECT count(*) FROM stocks').fetchone()
        assert result == (4,)
